# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import smtplib
import time
from abc import ABC
from abc import abstractmethod
from email.mime.text import MIMEText
from functools import reduce
from typing import List

from vdk.internal.builtin_plugins.notification.notification_configuration import (
    SmtpConfiguration,
)
from vdk.internal.core import errors

log = logging.getLogger(__name__)


class LoggingSMTP(smtplib.SMTP):
    """
    Subclass to replace the print statements with logging statements so that debug output can be captured
    """

    def _print_debug(self, *args):
        message = reduce((lambda msg1, msg2: str(msg1) + str(msg2)), iter(args))
        log.debug(message)


class INotification(ABC):
    @abstractmethod
    def notify(self, title, body):
        """
        Sends a message to recipients using a predefined channel.
        Both the receivers and the channel are specified in config.ini.
        :param title: message title
        :param body: message content
        """
        pass


class EmailNotification(INotification):
    def __init__(
        self,
        recipients: List[str],
        cc: List[str],
        smtp_cfg: SmtpConfiguration,
        sender: str,
    ):
        """
        :param recipients: list of emails
        :param sender: sender email
        :param cc: list of emails
        """
        log.debug(
            f"New EmailNotification prepared: sender: {sender}; cc: {cc}; recipeints: {recipients}"
        )
        self._sender = sender
        self._cc = [] if cc is None else cc
        self._smtp_cfg = smtp_cfg
        self._recipients = recipients

    def notify(self, subject, body):
        """
        Sends an email to the specified recipients. If there are no valid recipient addresses, it sends the email
        to the CC'd addresses, assuming at least one of them is valid. If none of them are, an appropriate error is
        logged, and no email is sent.

        :param subject: The subject of the email
        :param body: The email job
        """
        valid_recipients = ",".join(self._get_valid_recipients())
        if valid_recipients:
            email_message = self._build_message(subject, body, valid_recipients)
            self._send_message(email_message)
        else:
            valid_cc = ",".join(self._get_valid_cc())
            if valid_cc:
                logging.getLogger(__name__).warning(
                    "No valid recipient address is given"
                )
                email_message = self._build_message(
                    subject,
                    (
                        "This message is sent to the list of CC'd addresses "
                        "because none of the listed recipient addresses are valid, or none were provided.\n\n"
                    )
                    + body,
                    valid_cc,
                )
                self._send_message(email_message)
            elif (not valid_cc) and self._cc:
                logging.getLogger(__name__).warning(
                    "No valid recipients or CC addresses given"
                )
            else:
                logging.getLogger(__name__).debug("Empty list of recipients is given")

    def _send_message(self, msg):
        log.debug(f"Send message {msg}")
        s = self.__smtp_server()
        s.send_message(msg)
        s.quit()

    def _build_message(self, subject, body, recipients):
        msg = MIMEText(body, "html")
        msg["Subject"] = subject
        msg["From"] = self._sender
        msg["To"] = recipients
        msg["Cc"] = ",".join(self._get_valid_cc())
        return msg

    def _email_address_exists(self, email_address):
        """
        Checks if a particular email address is valid or not. The reason for this is that if an attempt to send
        an email to a non-valid address is made, that attempt will fail and no email will be sent to any of the
        specified addresses, regardless is any of the rest were valid or not.


        :param email_address: The email address in string form
        :return: a boolean specifying if the address is valid or not
        """
        s = self.__smtp_server()
        try:
            s.helo()
            s.mail(self._sender)
            resp = s.rcpt(email_address)
            return resp[0] == 250
        except smtplib.SMTPServerDisconnected as e:
            log.warning(
                "SMTP connection error while checking if email address exists: %s", e
            )
            return True
        finally:
            s.quit()

    def __smtp_server(self):
        s = LoggingSMTP(self._smtp_cfg.get_host(), port=self._smtp_cfg.get_port())
        if self._smtp_cfg.get_use_tls():
            s.starttls()
        if self._smtp_cfg.get_login_username():
            s.login(
                self._smtp_cfg.get_login_username(), self._smtp_cfg.get_login_password()
            )
        s.set_debuglevel(self._smtp_cfg.get_debug_level())
        return s

    def _get_valid_recipients(self):
        return [x.strip() for x in self._recipients if self._email_address_exists(x)]

    def _get_valid_cc(self):
        return [x.strip() for x in self._cc if self._email_address_exists(x)]


class EmailNotificationMessageBuilder(ABC):
    def __init__(self, job_name, job_log_url_template):
        self._job_name = job_name

        self.JOB_LOG_URL_TEMPLATE = job_log_url_template

        self.NOTIFICATION_MSG_TEMPLATE = """<p>Dear Versatile Data Kit user,<br />
   Last {exec_type} of your data job <strong>{job_name}</strong> has {result}.</p>
   {result_msg}
   <p>The sender mailbox is not monitored, please do not reply to this email.</p>
   <p>Best,<br />Versatile Data Kit team</p>"""

        self.USER_ERROR_TEMPLATE = """<p>Details:</p>
   <p style="padding-left: 30px; color:red;">{error_msg}</p>
   <p>{job_log_link}</p>
   <p>For more information please visit the user wiki.</p>"""

        self.INFRA_ERROR_MSG = """<p>Details:</p>
     <p style="padding-left: 30px; color:red;">{error_msg}</p>"""

        self.NOTIFICATION_MSG_SUBJECT_TEMPLATE = (
            "[{result}][data job {exec_type}] {job_name}"
        )

    @abstractmethod
    def build(self, exec_type, msg, op_id):
        """
        Builds a (subject, body) tuple to be used as input for EmailNotification.notify
        :param op_id: execution id of the data job
        :param msg: string to be included into the message template
        :param exec_type: 'run' or 'deploy'
        :return: (subject, body) tuple
        """
        pass

    def _get_job_log_url(self, op_id):
        now = int(round(time.time() * 1000))
        hour = 60 * 60 * 1000
        start_time_ms = now - 24 * hour
        end_time_ms = now + hour
        return self.JOB_LOG_URL_TEMPLATE.format(
            op_id=op_id, start_time_ms=start_time_ms, end_time_ms=end_time_ms
        )

    def _get_job_log_chunk(self, op_id):
        return (
            '<p>You can find full logs of the job execution <a href="{}"> here <a>.</p>'.format(
                self._get_job_log_url(op_id.strip())
            )
            if op_id
            else ""
        )

    def _get_subject(self, result, exec_type):
        return self.NOTIFICATION_MSG_SUBJECT_TEMPLATE.format(
            result=result, exec_type=exec_type, job_name=self._job_name
        )


class UserErrorEmailNotificationMessageBuilder(EmailNotificationMessageBuilder):
    def build(self, exec_type, msg="", op_id=""):
        user_error = self.USER_ERROR_TEMPLATE.format(
            error_msg=msg, job_log_link=self._get_job_log_chunk(op_id)
        )
        subject = self._get_subject("user error", exec_type)
        html_body = self.NOTIFICATION_MSG_TEMPLATE.format(
            exec_type=exec_type,
            job_name=self._job_name,
            result="failed",
            result_msg=user_error,
        )
        return subject, html_body


class InfraErrorEmailNotificationMessageBuilder(EmailNotificationMessageBuilder):
    def build(self, exec_type, msg="", op_id=""):
        if exec_type == "deploy":
            error_msg = errors.ErrorMessage(
                "There has been a platform error. The error will be resolved by the Versatile Data Kit team. "
                "Here are the details:",
                "Last deploy of your data job has failed",
                "There has been a platform error.",
                "Your new/updated job was not deployed. Your job will run its latest "
                "successfully deployed version (if any) as scheduled.",
                "The Versatile Data Kit team is working on the issue and will redeploy your job.",
            ).to_html()
        elif exec_type == "run":
            error_msg = errors.ErrorMessage(
                "There has been a platform error. The error will be resolved by the Versatile Data Kit team. "
                "Here are the details:",
                "Last run of your data job has failed",
                "There has been a platform error.",
                "Job will be restarted automatically on failure up to 3 times.",
                "No customer action is required. If the 3 restarts fail, Versatile Data Kit team will investigate the "
                "error and ensure job will continue to run as expected. If you want to receive notification on "
                "successful job execution, you can set it up in config.ini",
            ).to_html()
        else:
            logging.getLogger(__name__).warning(
                "Unknown exec_type [{}]".format(
                    exec_type if exec_type is not None else "None"
                )
            )
            exec_type = ""
            error_msg = "There has been a platform error. The error will be resolved by the Versatile Data Kit team."

        subject = self._get_subject("platform error", exec_type)
        html_body = self.NOTIFICATION_MSG_TEMPLATE.format(
            exec_type=exec_type,
            job_name=self._job_name,
            result="failed",
            result_msg=self.INFRA_ERROR_MSG.format(error_msg=error_msg),
        )
        return subject, html_body


class SuccessEmailNotificationMessageBuilder(EmailNotificationMessageBuilder):
    def build(self, exec_type, msg="", op_id=""):
        additional_info = self._get_job_log_chunk(op_id)
        subject = self._get_subject("success", exec_type)
        html_body = self.NOTIFICATION_MSG_TEMPLATE.format(
            exec_type=exec_type,
            job_name=self._job_name,
            result="succeeded",
            result_msg=additional_info,
        )
        return subject, html_body



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder

TERMINATION_MESSAGE_WRITER_ENABLED = "TERMINATION_MESSAGE_WRITER_ENABLED"
TERMINATION_MESSAGE_WRITER_OUTPUT_FILE = "TERMINATION_MESSAGE_WRITER_OUTPUT_FILE"


class WriterConfiguration:
    def __init__(self, config: Configuration):
        self.__config = config

    def get_writer_enabled(self):
        return bool(self.__config[TERMINATION_MESSAGE_WRITER_ENABLED])

    def get_output_file(self):
        return str(self.__config[TERMINATION_MESSAGE_WRITER_OUTPUT_FILE])


def add_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=TERMINATION_MESSAGE_WRITER_ENABLED,
        default_value=True,
        description="Set to false if you want to disable termination message writing.",
    )
    config_builder.add(
        key=TERMINATION_MESSAGE_WRITER_OUTPUT_FILE,
        default_value="/dev/termination-log",
        show_default_value=True,
        description="A file in which to write data job execution completion message.",
    )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

from vdk.api.job_input import IIngester
from vdk.api.plugin.plugin_input import IIngesterPlugin
from vdk.api.plugin.plugin_input import IIngesterRegistry
from vdk.internal.builtin_plugins.ingestion.ingester_base import IngesterBase
from vdk.internal.builtin_plugins.ingestion.ingester_configuration import (
    IngesterConfiguration,
)
from vdk.internal.builtin_plugins.run.execution_state import ExecutionStateStoreKeys
from vdk.internal.core import errors
from vdk.internal.core.config import Configuration
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import PlatformServiceError
from vdk.internal.core.errors import ResolvableBy
from vdk.internal.core.errors import UserCodeError
from vdk.internal.core.errors import VdkConfigurationError
from vdk.internal.core.statestore import CommonStoreKeys
from vdk.internal.core.statestore import StateStore
from vdk.internal.util.utils import parse_config_sequence


IngesterPluginFactory = Callable[[], IIngesterPlugin]

log = logging.getLogger(__name__)


class IngesterRouter(IIngesterRegistry, IIngester):
    """
    Router class for the core Ingestion API. It takes care of routing the
    payloads to their respective ingestion plugins.
    """

    def __init__(self, cfg: Configuration, core_state: StateStore):
        self._cfg: Configuration = cfg
        self._state: StateStore = core_state
        self._log: logging.Logger = logging.getLogger(__name__)
        self._cached_ingesters: Dict[str, IngesterBase] = dict()
        self._ingester_builders: Dict[str, IngesterPluginFactory] = dict()

    def add_ingester_factory_method(
        self,
        method: str,
        ingester_plugin: IngesterPluginFactory,
    ) -> None:
        """
        Add new ingester.
        """
        self._ingester_builders[method.lower()] = ingester_plugin

    def send_object_for_ingestion(
        self,
        payload: dict,
        destination_table: Optional[str] = None,
        method: Optional[str] = None,
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
    ):
        # Use the method and target provided by customer, or load the default ones
        # if set. `method` and `target` provided when the method is called take
        # precedence over the ones set with environment variables.
        method = method or self._cfg.get_value("INGEST_METHOD_DEFAULT")
        method = method.lower() if method is not None else None
        target = target or self._cfg.get_value("INGEST_TARGET_DEFAULT")
        self._log.info(
            "Sending object for ingestion with "
            f"method: {method} and target: {target}"
        )

        if method in self._cached_ingesters.keys():
            ingester = self._cached_ingesters.get(method)
            self.__ingest_object(
                ingester=ingester,
                payload=payload,
                destination_table=destination_table,
                method=method,
                target=target,
                collection_id=collection_id,
            )

        elif method in self._ingester_builders.keys():
            ingester = self.__initialize_ingester(method=method)
            self.__ingest_object(
                ingester=ingester,
                payload=payload,
                destination_table=destination_table,
                method=method,
                target=target,
                collection_id=collection_id,
            )
        else:
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=self._log,
                what_happened=f"Provided method, {method}, has invalid value.",
                why_it_happened=f"VDK was run with method={method}, however {method} is not part of the available ingestion mechanisms.",
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures=f"Provide either valid value for method, or install ingestion plugin that supports this type. "
                f"Currently possible values are {list(self._ingester_builders.keys())}",
            )

    def send_tabular_data_for_ingestion(
        self,
        rows: iter,
        column_names: list,
        destination_table: Optional[str] = None,
        method: Optional[str] = None,
        target: Optional[str] = None,
        collection_id: Optional[str] = None,
    ):
        # Use the method and target provided by customer, or load the
        # default ones, if set. `method` and `target` provided when the
        # method is called take precedence over the ones set with
        # environment variables.
        method = method or self._cfg.get_value("INGEST_METHOD_DEFAULT")
        target = target or self._cfg.get_value("INGEST_TARGET_DEFAULT")
        self._log.info(
            "Sending tabular data for ingestion with "
            f"method: {method} and target: {target}"
        )

        if method in self._cached_ingesters.keys():
            ingester = self._cached_ingesters.get(method)
            self.__ingest_tabular_data(
                ingester=ingester,
                rows=rows,
                column_names=column_names,
                destination_table=destination_table,
                method=method,
                target=target,
                collection_id=collection_id,
            )

        elif method in self._ingester_builders.keys():
            ingester = self.__initialize_ingester(method=method)
            self.__ingest_tabular_data(
                ingester=ingester,
                rows=rows,
                column_names=column_names,
                destination_table=destination_table,
                method=method,
                target=target,
                collection_id=collection_id,
            )
        else:
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=self._log,
                what_happened=f"Provided method, {method}, has invalid value.",
                why_it_happened=f"VDK was run with method={method}, however {method} is not part of the available ingestion mechanisms.",
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures=f"Provide either valid value for method, or install ingestion plugin that supports this type. "
                f"Currently possible values are {list(self._ingester_builders.keys())}",
            )

    def __ingest_object(
        self,
        ingester: IngesterBase,
        payload: dict,
        destination_table: Optional[str],
        method: Optional[str],
        target: Optional[str],
        collection_id: Optional[str],
    ):
        try:
            ingester.send_object_for_ingestion(
                payload, destination_table, method, target, collection_id
            )
        except Exception as e:
            self._log.error(
                "Failed to send object for ingestion." f"Exception was: {e}"
            )

    def __ingest_tabular_data(
        self,
        ingester: IngesterBase,
        rows: iter,
        column_names: iter,
        destination_table: Optional[str],
        method: Optional[str],
        target: Optional[str],
        collection_id: Optional[str],
    ):
        try:
            ingester.send_tabular_data_for_ingestion(
                rows, column_names, destination_table, method, target, collection_id
            )
        except Exception as e:
            self._log.error(
                "Failed to send tabular data for ingestion." f"Exception was: {e}"
            )
            errors.log_and_rethrow(
                ResolvableBy.USER_ERROR,
                self._log,
                what_happened="Failed to send tabular data for ingestion",
                why_it_happened=f"Exception was: {e}",
                consequences="Data is not ingested and the method will raise an exception",
                countermeasures="Please look the error message and try to fix the error and re-try.",
                exception=e,
                wrap_in_vdk_error=True,
            )

    def __initialize_ingester(self, method) -> IngesterBase:
        ingester_plugin = None

        try:
            ingester_plugin = self._ingester_builders[method]()
        except KeyError:
            self._log.error("Could not initialize ingestion plugin.")

        if ingester_plugin is None:
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.CONFIG_ERROR,
                log=self._log,
                what_happened=f"Could not create new ingester plugin of type {method}.",
                why_it_happened=f"VDK was run with method={method}, however no valid ingester plugin was created.",
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures=f"Seems to be a bug in the plugin for method {method}. Make sure it's correctly installed. "
                f"If upgraded recently consider reverting to previous version. Or use another method type.",
            )
        else:
            # Initialize the pre- and post- processors.
            initialized_pre_processors = self.__get_initialized_processors(
                "INGEST_PAYLOAD_PREPROCESS_SEQUENCE"
            )
            initialized_post_processors = self.__get_initialized_processors(
                "INGEST_PAYLOAD_POSTPROCESS_SEQUENCE"
            )

            self._cached_ingesters[method] = IngesterBase(
                self._state.get(ExecutionStateStoreKeys.JOB_NAME),
                self._state.get(CommonStoreKeys.OP_ID),
                ingester_plugin,
                IngesterConfiguration(config=self._cfg),
                pre_processors=initialized_pre_processors,
                post_processors=initialized_post_processors,
            )

        return self._cached_ingesters[method]

    def __get_initialized_processors(self, config_var: str) -> List:
        return [
            self.__initialize_processor(i)
            for i in parse_config_sequence(self._cfg, key=config_var, sep=",")
        ]

    def __initialize_processor(self, method) -> Optional[IIngesterPlugin]:
        processor_plugin = None

        try:
            processor_plugin = self._ingester_builders[method]()
        except KeyError:
            self._log.error("Could not initialize processor plugin.")

        if processor_plugin is None:
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.CONFIG_ERROR,
                log=self._log,
                what_happened="Could not create new processor plugin of type"
                f" {method}.",
                why_it_happened=f"VDK was run with method={method}, however "
                "no valid ingestion processor plugin was "
                "created.",
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures="Seems to be a bug in the plugin for method"
                f" {method}. Make sure it's correctly "
                f"installed. If upgraded recently, consider"
                " reverting to previous version. Or use "
                "another method type.",
            )
        return processor_plugin

    def close(self):
        """
        Wait for all ingestion payloads to be sent and clean up the queues. This
        method iterates over all ingestion plugins, and blocks the termination of the
        process before all queues are emptied.
        """
        errors_list: Dict = dict()
        for method, ingester in self._cached_ingesters.items():
            try:
                ingester.close()
            except Exception as e:
                errors_list[method] = e
                self._log.warning(
                    f"Closing of ingester for method `{method}` failed. "
                    f"Exception was: {e}"
                )

        if errors_list:
            message = ErrorMessage(
                "Ingesting data failed",
                f"On close some following ingest queues types reported errors:  {list(errors_list.keys())}.",
                f"There were errors while closing ingestion. Exceptions were: {errors_list}.",
                "Some data was partially ingested or not ingested at all.",
                "Follow the instructions in the error messages and log warnings. "
                "Make sure to inspect any errors or warning logs generated"
                "Re-try the job if necessary",
            )

            if any(
                filter(lambda v: isinstance(v, UserCodeError), errors_list.values())
            ):
                raise UserCodeError(message)
            elif any(
                filter(
                    lambda v: isinstance(v, VdkConfigurationError), errors_list.values()
                )
            ):
                raise VdkConfigurationError(message)
            else:
                raise PlatformServiceError(message)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from typing import List

from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.internal.core.config import Configuration
from vdk.internal.core.statestore import ImmutableStoreKey
from vdk.internal.core.statestore import StateStore
from vdk.internal.core.statestore import StoreKey


# TODO: consider moving some of those (or extracting interface) to vdk.api as they are used by users (plugin
#  developers)


@dataclass(frozen=True)
class CoreContext:
    """
    The context is a special object that hold info about the context in which the current execution is.

    It keeps info about the plugins being loaded and enable manipulating plugins
    It keeps info about how is the current execution being configured.
    It also holds state relevant for the CLI execution.
    """

    plugin_registry: IPluginRegistry
    configuration: Configuration
    # Keep state here, It's mutable and can be edited by plugins as well.
    # Key starting with "vdk." should be reserved for core state and not used by plugins
    # For data job context (run command)
    # there's separate JobContext class that's should be used (see builtin_plugins/run/job_context)
    state: StateStore

    def create_child_context(self):
        """
        You can create a child context if you want to create internal child execution of a CLI execution.
        For example CLI (root level) execution will keep track of generic CLI state.
        Data Job execution being spawned in CLI execution would be a child of that CLI execution and keep track of
        Job execution related state.
        Child context inherits (copies) parent context at the time of creation. Permanent link to parent is not tracked.
        It's kept as key only for reference purposes.
        """

        child_context = CoreContext(
            self.plugin_registry, self.configuration, self.state.clone()
        )
        child_context.state.set(StoreKey[StateStore]("vdk.state.parent"), self.state)
        children_key = ImmutableStoreKey[List[StateStore]]("vdk.state.children")
        if self.state.get(children_key) is None:
            self.state.set(children_key, [])
        self.state.get(children_key).append(child_context.state)
        return child_context



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
from typing import List
from typing import Tuple

from vdk.api.plugin import hook_markers
from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.api.plugin.plugin_registry import PluginException
from vdk.api.plugin.plugin_registry import PluginHookRelay
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import ResolvableBy
from vdk.internal.plugin.plugin_manager import VdkPluginManager
from vdk.internal.util.utils import log_plugin_load_fail


log = logging.getLogger(__name__)


class PluginRegistry(IPluginRegistry):
    """
    Manage plugins.
    Plugin consist of two things
    * Specification (see hookspec decorator) which define method (name and arguments)
    * Implementation (see hookimpl decorator) which must matches method name and subset of the arguments

    Plugin Registry will automatically load all plugins registered through setup tools entry points for vdk.plugin.run
    And provide methods for adding new specs and implementations.

    see https://pluggy.readthedocs.io/en/latest/#how-does-it-work
    """

    def __init__(self, group_name=hook_markers.GROUP_NAME, plugin_manager=None):
        self.__plugin_manager = (
            VdkPluginManager(group_name) if plugin_manager is None else plugin_manager
        )
        self.__hook_relay = PluginHookRelay(self.__plugin_manager.hook)
        self.__group_name = group_name
        self.__plugin_load_success = False

    def __str__(self):
        s = (
            f"PluginRegistry(group_name={self.__group_name})"
            f"Plugins: " + f"{os.linesep}".join([str(v) for v in self.list_plugins()])
        )
        return s

    def plugin_manager(self):
        return self.__plugin_manager

    def load_plugins_from_setuptools_entrypoints(self) -> None:
        log.debug("Loading plugins")

        try:
            self.__plugin_manager.load_setuptools_entrypoints(self.__group_name)
        except ImportError as e:
            log_plugin_load_fail(ResolvableBy.USER_ERROR, log, e, self.__group_name)
            raise PluginException from e
        except Exception as e:
            log_plugin_load_fail(ResolvableBy.PLATFORM_ERROR, log, e, self.__group_name)
            raise PluginException from e

        plugins = self.__plugin_manager.list_name_plugin()
        log.info(
            f"Following plugins from setup entrypoints have been discovered and registered: {plugins}"
        )
        self.__plugin_load_success = True

    def is_plugin_load_success(self):
        return self.__plugin_load_success

    def list_plugins(self) -> List[Tuple[str, str]]:
        return self.__plugin_manager.list_name_plugin()

    def has_plugin(self, name: str) -> bool:
        return self.__plugin_manager.has_plugin(name)

    def load_plugin_with_hooks_impl(
        self, module_or_class_with_hook_impls: object, name: str = None
    ) -> None:
        """
        Load new plugin with hook implementations

        :param module_or_class_with_hook_impls: can be module (python package) or class or class instance
        :param name: the name of the plugin. If None it will be the cannonical name of the module/class passed.
        which contain the hook implementations
        """
        try:
            plugin_name = self.__plugin_manager.register(
                module_or_class_with_hook_impls, name=name
            )
            if plugin_name:
                log.debug(f"Registered new plugin: {plugin_name}")
            else:
                log.warning(
                    f"Failed to register plugin {name}. Most likely the plugin name has been forbidden"
                )
        except Exception as e:
            message = ErrorMessage(
                summary=f"Failed to load plugin ",
                what=f"Failed to load plugin with name  '{name}' and module/class '{module_or_class_with_hook_impls}'",
                why="Most likely a bug in the plugin",
                consequences="The CLI tool will likely abort.",
                countermeasures="Re-try again. "
                "Revert to previous stable version of the plugin or CLI "
                "plugin (pip install vdk-plugin-name==version) "
                "Or see what plugins are installed (use `pip list` command) and if "
                "there are not issues. "
                "Or try to reinstall the app in a new clean environment  ",
            )
            raise PluginException(message) from e

    def add_hook_specs(self, module_or_class_with_hookspecs: object):
        self.__plugin_manager.add_hookspecs(module_or_class_with_hookspecs)

    def hook(self) -> PluginHookRelay:
        """
        Hook relay which to use to execute the plugin hooks with.

        You can also cast it to the HookSpec class with the hook you want to invoke to get auto-completion and so on.
        results: List = cast(CoreHookSpecs, plugin_registry.hook()).some_hook()
        But NOTE that the return result is a list of the results of all hooks implementations
        (unless hook spec is firstresult=True)
        """
        return self.__hook_relay



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
import math
import os
import pathlib
import re
import sys
from typing import cast
from typing import Dict
from typing import List
from typing import Optional

import click
from vdk.internal.builtin_plugins.config import vdk_config
from vdk.internal.builtin_plugins.config.job_config import JobConfig
from vdk.internal.builtin_plugins.run import job_input_error_classifier
from vdk.internal.builtin_plugins.run.data_job import DataJobFactory
from vdk.internal.builtin_plugins.run.execution_tracking import (
    ExecutionTrackingPlugin,
)
from vdk.internal.builtin_plugins.version import version
from vdk.internal.core import errors
from vdk.internal.core.context import CoreContext
from vdk.internal.core.errors import VdkConfigurationError

log = logging.getLogger(__name__)


class CliRunImpl:
    def __init__(self, job_factory: DataJobFactory = None):
        self.__job_factory = job_factory if job_factory else DataJobFactory()

    @staticmethod
    def __validate_and_parse_args(arguments: str) -> Optional[Dict]:
        try:
            if arguments:
                return json.loads(arguments)
            else:
                return None
        except Exception as e:
            blamee = errors.ResolvableBy.USER_ERROR
            errors.log_and_rethrow(
                blamee,
                logging.getLogger(__name__),
                what_happened="Failed to validate job arguments.",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                consequences=errors.MSG_CONSEQUENCE_TERMINATING_APP,
                countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION,
                exception=e,
                wrap_in_vdk_error=True,
            )

    @staticmethod
    def __split_into_chunks(exec_steps: List, chunks: int) -> List:
        """
        Generator that splits the list of execution steps into sequential
        sub-lists.
        :param exec_steps: the list of execution steps
        :param chunks: the number of sublists that need to be produced
        :return: a list of the sub-lists produced
        """
        quotient, remainder = divmod(len(exec_steps), chunks)
        for i in range(chunks):
            subsequent_iteration = (quotient + 1) * (
                i if i < remainder else remainder
            ) + quotient * (0 if i < remainder else i - remainder)
            yield exec_steps[
                subsequent_iteration : subsequent_iteration
                + (quotient + 1 if i < remainder else quotient)
            ]

    @staticmethod
    def __warn_on_python_version_disparity(
        context: CoreContext, job_directory: pathlib.Path
    ):
        log_config_type = context.configuration.get_value(vdk_config.LOG_CONFIG)
        if log_config_type == "LOCAL":
            # Get the local python installation's version.
            python_env_version = sys.version_info
            local_py_version = f"{python_env_version.major}.{python_env_version.minor}"

            # Get the python_version set in the config.ini if any.
            job_path = job_directory.resolve()
            try:
                config = JobConfig(data_job_path=job_path)
            except VdkConfigurationError as e:
                log.info(
                    f"An exception occurred while loading job configuration. Error was {e}"
                )
                return
            configured_python_version = config.get_python_version()

            if not configured_python_version:
                return

            pattern = r"^\d+\.\d+"
            version_match = re.match(pattern, configured_python_version)

            if version_match:
                extracted_configured_version = version_match.group()
                if extracted_configured_version != local_py_version:
                    log.warning(
                        f"""
                        {os.linesep + (' ' * 20) + ('*' * 80)}
                        What: Python version mismatch between local python and configure python.
                        Why: The Python version specified in the job's config.ini file  is ({configured_python_version})
                        while the local python version used to execute the data job is ({local_py_version}).
                        Consequences: Developing a data job using one Python version and deploying
                        it with a different version can result in unexpected and
                        difficult-to-troubleshoot errors like module incompatibilities, or
                        unexpected behavior during execution.
                        Countermeasures: To resolve this issue, ensure that the Python version
                        specified in the python_version property of the config.ini file matches
                        the Python version of your execution environment by either editing the
                        python_version property in config.ini, or switching local environment
                        to a matching version of python.
                        {os.linesep + (' ' * 20) + ('*' * 80)}
                        """
                    )

    def create_and_run_data_job(
        self,
        context: CoreContext,
        data_job_directory: pathlib.Path,
        arguments: Optional[str],
    ):
        log.info(f"Run job with directory {data_job_directory}")
        context.plugin_registry.load_plugin_with_hooks_impl(ExecutionTrackingPlugin())

        self.__warn_on_python_version_disparity(
            context=context, job_directory=data_job_directory
        )

        job = self.__job_factory.new_datajob(
            data_job_directory=data_job_directory, core_context=context
        )
        args = self.__validate_and_parse_args(arguments)

        execution_result = None
        try:
            execution_result = job.run(args)

            # On some platforms, if the size of a string is too large, the
            # logging module starts throwing OSError: [Errno 40] Message too long,
            # so it is safer if we split large strings into smaller chunks.
            string_exec_result = str(execution_result)
            if len(string_exec_result) > 5000:
                temp_exec_result = json.loads(string_exec_result)
                steps = temp_exec_result.pop("steps_list")

                log.info(
                    f"Data Job execution summary: {json.dumps(temp_exec_result, indent=2)}"
                )

                chunks = math.ceil(len(string_exec_result) / 5000)
                for i in self.__split_into_chunks(exec_steps=steps, chunks=chunks):
                    log.info(f"Execution Steps: {json.dumps(i, indent=2)}")

            else:
                log.info(f"Data Job execution summary: {execution_result}")
        except BaseException as e:
            errors.log_and_rethrow(
                job_input_error_classifier.whom_to_blame(
                    e, __file__, data_job_directory
                ),
                log,
                what_happened="Failed executing job.",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                consequences=errors.MSG_CONSEQUENCE_TERMINATING_APP,
                countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION
                + " Most likely a prerequisite or plugin of one of the key VDK components failed, see"
                + " logs for details and ensure the prerequisite for the failed component (details in stacktrace).",
                exception=e,
            )
        if execution_result.is_failed() and execution_result.get_exception_to_raise():
            raise execution_result.get_exception_to_raise()


@click.command(
    help="Run a Data Job. "
    """
     Examples:

     \b
     # This will run the Data Job from directory example-job (it takes a few minutes)
     vdk run /home/user/data-jobs/example-job
     \b
     # Run data job with arguments.
     vdk run example-job --arguments '{"key1": "value1","key2": "value2"}'

"""
)
@click.argument(
    "data_job_directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
)
@click.option(
    "--arguments",
    type=click.STRING,
    required=False,
    help="Pass arguments. "
    "Those arguments will be passed to each step. "
    "Must be in valid JSON format. "
    "Arguments are passed to each step. "
    "They can be used as parameters in SQL queries and will be replaced automatically."
    "Properties can also be specified as parameters in SQL, arguments would have higher priority for the same key.",
)
@click.pass_context
def run(ctx: click.Context, data_job_directory: str, arguments: str) -> None:
    """
    Entry point of the CLI run. It start a run (execution) of a data job.
    """
    log.info(
        f"Versatile Data Kit (VDK){os.linesep}{version.get_version_info()}{os.linesep + '-' * 80}"
    )
    context: CoreContext = cast(CoreContext, ctx.obj)
    run_impl = CliRunImpl()
    run_impl.create_and_run_data_job(
        context, pathlib.Path(data_job_directory), arguments
    )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import re
import socket
import types
from sys import modules
from typing import cast

from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.config import vdk_config
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core import errors
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.errors import ResolvableBy
from vdk.internal.core.statestore import CommonStoreKeys

SYSLOG_URL = "SYSLOG_URL"
SYSLOG_PORT = "SYSLOG_PORT"
SYSLOG_SOCK_TYPE = "SYSLOG_SOCK_TYPE"
SYSLOG_ENABLED = "SYSLOG_ENABLED"

SYSLOG_SOCK_TYPE_VALUES_DICT = {"UDP": socket.SOCK_DGRAM, "TCP": socket.SOCK_STREAM}


def _parse_log_level_module(log_level_module):
    valid_logging_levels = [
        "NOTSET",
        "DEBUG",
        "INFO",
        "WARN",
        "WARNING",
        "ERROR",
        "FATAL",
        "CRITICAL",
    ]
    try:
        if log_level_module and log_level_module.strip():
            modules = log_level_module.split(";")
            result = {}
            for module in modules:
                if module:
                    module_and_level = module.split("=")
                    if not re.search("[a-zA-Z0-9_.-]+", module_and_level[0].lower()):
                        raise ValueError(
                            f"Invalid logging module name: '{module_and_level[0]}'. "
                            f"Must be alphanumerical/underscore characters."
                        )
                    if module_and_level[1].upper() not in valid_logging_levels:
                        raise ValueError(
                            f"Invalid logging level: '{module_and_level[1]}'. Must be one of {valid_logging_levels}."
                        )
                    result[module_and_level[0]] = {"level": module_and_level[1].upper()}
            return result
        else:
            return {}
    except Exception as e:
        errors.log_and_throw(
            ResolvableBy.CONFIG_ERROR,
            logging.getLogger(__name__),
            "Invalid logging configuration passed to LOG_LEVEL_MODULE.",
            f"Error is: {e}. log_level_module was set to {log_level_module}.",
            "Logging will not be initialized and exception is raised",
            "Set correctly configuration to log_level_debug configuration in format 'module=level;module2=level2'",
        )


def configure_loggers(
    job_name: str = "",
    attempt_id: str = "no-id",
    log_config_type: str = None,
    vdk_logging_level: str = "DEBUG",
    log_level_module: str = None,
    syslog_args: (str, int, str, bool) = ("localhost", 514, "UDP", False),
) -> None:
    """
    Configure default logging configuration

    :param job_name: the name of the job
    :param attempt_id: the id of the current job run attempt
    :param log_config_type: where the job is executed: CLOUD or LOCAL
    :param vdk_logging_level: The level for vdk specific logs.
    :param log_level_module: The level for modules specific logs
    :param syslog_args: Arguments necessary for SysLog logging.
    """

    import logging.config

    if _is_already_configured():
        log = logging.getLogger(__name__)
        msg = "Logging already configured. This seems like a bug. Fix me!."
        # PLugins would override loggin directly wiht logging library not through this method
        # it is probably OK for tests to configure logging twice. VDK, however, shouldn't.
        log.warning(msg)

    # Likely most logging system (like Log Insight) show the hostname from where the syslog message arrived so no need to include it here.
    DETAILED_FORMAT = (
        f"%(asctime)s [VDK] {job_name} [%(levelname)-5.5s] %(name)-30.30s %(filename)20.20s:%("
        f"lineno)-4.4s %(funcName)-16.16s[id:{attempt_id}]- %(message)s"
    )

    _LOGGERS = {
        "requests_kerberos": {"level": "INFO"},
        "requests_oauthlib": {"level": "INFO"},
        "urllib3": {"level": "INFO"},
        "vdk": {"level": vdk_logging_level},
    }
    _LOGGERS.update(_parse_log_level_module(log_level_module))

    _FORMATTERS = {"detailedFormatter": {"format": DETAILED_FORMAT}}

    _CONSOLE_HANDLER = {
        "class": "logging.StreamHandler",
        "formatter": "detailedFormatter",
        "stream": "ext://sys.stderr",
    }

    syslog_url, syslog_port, syslog_sock_type, syslog_enabled = syslog_args

    if syslog_sock_type not in SYSLOG_SOCK_TYPE_VALUES_DICT:
        errors.log_and_throw(
            to_be_fixed_by=errors.ResolvableBy.CONFIG_ERROR,
            log=log,
            what_happened=f"Provided configuration variable for {SYSLOG_SOCK_TYPE} has invalid value.",
            why_it_happened=f"VDK was run with {SYSLOG_SOCK_TYPE}={syslog_sock_type}, however {syslog_sock_type} is invalid value for this variable.",
            consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
            countermeasures=f"Provide a valid value for {SYSLOG_SOCK_TYPE}."
            f"Currently possible values are {list(SYSLOG_SOCK_TYPE_VALUES_DICT.keys())}",
        )

    _SYSLOG_HANDLER = {
        "level": "DEBUG",
        "class": "logging.handlers.SysLogHandler",
        "formatter": "detailedFormatter",
        "address": (syslog_url, syslog_port),
        "socktype": SYSLOG_SOCK_TYPE_VALUES_DICT[syslog_sock_type],
        "facility": "user",
    }

    handlers = ["consoleHandler"]
    if syslog_enabled:
        handlers.append("SysLog")

    if "CLOUD" == log_config_type:
        CLOUD = {  # @UnusedVariable
            "version": 1,
            "handlers": {"consoleHandler": _CONSOLE_HANDLER, "SysLog": _SYSLOG_HANDLER},
            "formatters": _FORMATTERS,
            "root": {"handlers": handlers},
            "loggers": _LOGGERS,
            "disable_existing_loggers": False,
        }
        logging.config.dictConfig(CLOUD)
    elif "NONE" == log_config_type:
        pass
    else:
        LOCAL = {  # @UnusedVariable
            "version": 1,
            "handlers": {"consoleHandler": _CONSOLE_HANDLER, "SysLog": _SYSLOG_HANDLER},
            "formatters": _FORMATTERS,
            "root": {"handlers": handlers},
            "loggers": _LOGGERS,
            "disable_existing_loggers": False,
        }
        logging.config.dictConfig(LOCAL)
    _set_already_configured()


def _set_already_configured() -> None:
    setattr(modules[__name__], "logging_already_configured", True)


def _is_already_configured() -> bool:
    res = hasattr(modules[__name__], "logging_already_configured")
    return res


def ensure_logging_for_current_test() -> None:
    """
    To be called from test code only.

    Ensures that logging is configured, either by another class, or by the caller.
    Useful when debugging a single test class, not the whole suite (assuming all suites configure logging)

    Uses filename of caller for derived_dataset parameter, e.g. if this method is called from impala_connection_test.py
    then logs would be configured to look like:

    2018-04-19 16:05:48,488=1524143148[VDK] impala_connection_test.py [DEBUG]  ......
    """
    if not _is_already_configured():
        try:
            from inspect import currentframe

            me = cast(types.FrameType, currentframe())
            caller = cast(types.FrameType, me.f_back)
            path = caller.f_locals["__file__"]
            from os.path import split

            dir, fname = split(path)  # @ReservedAssignment
        except:
            fname = "n/a"
        from time import time

        configure_loggers(fname, str(time()))


class LoggingPlugin:
    """
    Define the logging plugin
    """

    @staticmethod
    @hookimpl
    def vdk_configure(config_builder: ConfigurationBuilder):
        config_builder.add(
            key=SYSLOG_URL,
            default_value="localhost",
            description="The hostname of the endpoint to which VDK logs will be sent through SysLog.",
        )
        config_builder.add(
            key=SYSLOG_PORT,
            default_value=514,
            description="The port of the endpoint to which VDK logs will be sent through SysLog.",
        )
        config_builder.add(
            key=SYSLOG_ENABLED,
            default_value=False,
            description="If set to True, SysLog log forwarding is enabled.",
        )
        config_builder.add(
            key=SYSLOG_SOCK_TYPE,
            default_value="UDP",
            description=f"Socket type for SysLog log forwarding connection. Currently possible values are {list(SYSLOG_SOCK_TYPE_VALUES_DICT.keys())}",
        )

    @hookimpl
    def initialize_job(self, context: JobContext) -> None:
        """
        Initialize logging for running Data Job.

        When running a data job logging is more verbose, it includes more details like op id, job name.
        """

        attempt_id = context.core_context.state.get(CommonStoreKeys.ATTEMPT_ID)
        job_name = context.name
        log_config_type = context.core_context.configuration.get_value(
            vdk_config.LOG_CONFIG
        )
        vdk_log_level = context.core_context.configuration.get_value(
            vdk_config.LOG_LEVEL_VDK
        )
        log_level_module = context.core_context.configuration.get_value(
            vdk_config.LOG_LEVEL_MODULE
        )
        syslog_url = context.core_context.configuration.get_value(SYSLOG_URL)
        syslog_port = context.core_context.configuration.get_value(SYSLOG_PORT)
        syslog_sock_type = context.core_context.configuration.get_value(
            SYSLOG_SOCK_TYPE
        )
        syslog_enabled = context.core_context.configuration.get_value(SYSLOG_ENABLED)
        try:  # If logging initialization fails we want to attempt sending telemetry before exiting VDK
            configure_loggers(
                job_name,
                attempt_id,
                log_config_type=log_config_type,
                vdk_logging_level=vdk_log_level,
                log_level_module=log_level_module,
                syslog_args=(syslog_url, syslog_port, syslog_sock_type, syslog_enabled),
            )
            log = logging.getLogger(__name__)
            log.debug(f"Initialized logging for log type {log_config_type}.")
        except Exception as e:
            # Have local logs on DEBUG level, when standard log configuration fails
            logging.basicConfig(level=logging.DEBUG)
            log = logging.getLogger(__name__)
            log.warning(
                "Logging configuration %s failed.\n"
                "Trying to send telemetry for Job attempt: %s"
                % (log_config_type, attempt_id)
            )
            errors.log_and_rethrow(
                errors.find_whom_to_blame_from_exception(e),
                log=logging.getLogger(__name__),
                what_happened="Failed to initialize logging",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                consequences="Logging is critical VDK component. VDK will now exit.",
                countermeasures="Depending on stacktrace",
                exception=e,
            )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pluggy
from vdk.api.plugin.hook_markers import hookimpl
from vdk.api.plugin.plugin_registry import HookCallResult
from vdk.internal.builtin_plugins.config import vdk_config
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.builtin_plugins.run.execution_results import StepResult
from vdk.internal.builtin_plugins.run.execution_state import ExecutionStateStoreKeys
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.builtin_plugins.run.run_status import ExecutionStatus
from vdk.internal.builtin_plugins.run.step import Step


class ExecutionTrackingPlugin:
    """
    We are going to track data job executions in this plugin and update the statestore of the execution.
    """

    @hookimpl
    def initialize_job(self, context: JobContext):
        state = context.core_context.state
        configuration = context.core_context.configuration

        state.set(ExecutionStateStoreKeys.JOB_NAME, context.name)
        state.set(
            ExecutionStateStoreKeys.JOB_GIT_HASH,
            configuration.get_value(vdk_config.JOB_GITHASH),
        )
        state.set(ExecutionStateStoreKeys.STEPS_STARTED, [])
        state.set(ExecutionStateStoreKeys.STEPS_SUCCEEDED, [])
        state.set(ExecutionStateStoreKeys.STEPS_FAILED, [])
        state.set(ExecutionStateStoreKeys.EXECUTION_RESULT, None)

    @hookimpl(hookwrapper=True)
    def run_job(self, context: JobContext) -> None:
        """The script that runs the actual run of the data job.
        It executes the provided steps starting from context.step_tree_root in sequential order (using BFS)
        using
        """
        out: HookCallResult
        out = yield

        result: ExecutionResult = out.get_result()
        state = context.core_context.state
        state.set(ExecutionStateStoreKeys.EXECUTION_RESULT, result)

    @hookimpl(hookwrapper=True)
    def run_step(self, context: JobContext, step: Step) -> None:
        state = context.core_context.state

        state.get(ExecutionStateStoreKeys.STEPS_STARTED).append(step.name)
        out: HookCallResult
        out = yield
        if out.excinfo:
            state.get(ExecutionStateStoreKeys.STEPS_FAILED).append(step.name)

        result: StepResult = out.get_result()  # will throw if there was an exception
        if result.status == ExecutionStatus.SUCCESS:
            state.get(ExecutionStateStoreKeys.STEPS_SUCCEEDED).append(step.name)
        elif result.status == ExecutionStatus.ERROR:
            state.get(ExecutionStateStoreKeys.STEPS_FAILED).append(step.name)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from copy import deepcopy
from typing import Dict

from vdk.api.plugin.plugin_input import ISecretsServiceClient

log = logging.getLogger(__name__)


class InMemSecretsServiceClient(ISecretsServiceClient):
    """
    Implementation of IProperties that are kept only in memory.
    """

    def __init__(self):
        self._secrets = {}

    def read_secrets(self, job_name: str, team_name: str) -> Dict:
        res = deepcopy(self._secrets)
        return res

    def write_secrets(self, job_name: str, team_name: str, secrets: Dict) -> Dict:
        log.warning(
            "You are using In Memory Secrets client. "
            "That means the secrets will not be persisted past the Data Job run."
        )
        self._secrets = deepcopy(secrets)
        return self._secrets



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import pathlib
import tempfile
from os import getenv
from typing import Optional

from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.config.job_config import JobConfig
from vdk.internal.builtin_plugins.config.job_config import JobConfigKeys
from vdk.internal.core.config import ConfigurationBuilder

VDK_ = "VDK_"

# configuration keys:
DB_DEFAULT_TYPE = "DB_DEFAULT_TYPE"
JOB_GITHASH = "JOB_GITHASH"
LOG_CONFIG = "LOG_CONFIG"
LOG_LEVEL_VDK = "LOG_LEVEL_VDK"
LOG_LEVEL_MODULE = "LOG_LEVEL_MODULE"
WORKING_DIR = "WORKING_DIR"
ATTEMPT_ID = "ATTEMPT_ID"
EXECUTION_ID = "EXECUTION_ID"
OP_ID = "OP_ID"
TEMPORARY_WRITE_DIRECTORY = "TEMPORARY_WRITE_DIRECTORY"
PYTHON_VERSION = "PYTHON_VERSION"

log = logging.getLogger(__name__)


class CoreConfigDefinitionPlugin:
    """
    Define the core configuration.

    The configuration defaults to behavior needed by data engineers/analysts
       (i.e. running locally against production with no monitoring and console logging)
    Automated execution in cloud should override:
       1. VDK_MONITORING_ENABLED = True
       2. VDK_LOG_CONFIG = CLOUD
       4. VDK_WORKING_DIR
    """

    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        """Set common configuration necessary for vdk

        We are setting it as tryfirst to make sure any config provider afterwards would see the definitions.
        """

        # TODO: Get currently supported db types
        db_default_type_description = (
            "Default DB connection provided by VDK when executing a data job. "
            "All sql queries, templates/loads would be against that database."
            "Different database types can be configured with plugins. \n"
        )
        config_builder.add(DB_DEFAULT_TYPE, None, True, db_default_type_description)

        config_builder.add(
            WORKING_DIR, "."
        )  # Use '.' when running on the cloud. User ssh-admin has rights to current folder

        config_builder.add("MONITORING_ENABLED", False)
        config_builder.add(
            LOG_CONFIG, "LOCAL"
        )  # {LOCAL, CLOUD, NONE} To be overridden when executing in cloud
        config_builder.add(
            LOG_LEVEL_VDK,
            None,
            True,
            "Logging verbosity of VDK code can be controlled from here. "
            "Allowed values: CRITICAL, ERROR, WARNING, INFO, DEBUG. "
            "If not set python default or one set by vdk -v LEVEL is used. ",
        )
        config_builder.add(
            LOG_LEVEL_MODULE,
            None,
            True,
            "Logging verbosity of specific module can be controlled from here."
            "Format is semi colon separated list of module and level."
            " module=level,module2=level2 For example a.b.c=INFO;foo.bar=ERROR "
            "Allowed values: CRITICAL, ERROR, WARNING, INFO, DEBUG. "
            "If not set python default or one set by vdk -v LEVEL is used. ",
        )
        config_builder.add(JOB_GITHASH, "unknown")
        config_builder.add(
            OP_ID,
            "",
            True,
            "It identifies the trigger that initiated this job. "
            "If left empty it will be auto-generated."
            "OP ID is similar to trace ID in open tracing."
            "It enable tracing multiple operations across difference services and datasets"
            "For example, it is possible to have N jobs with same OpID (if Job1 started Job2 then Job1.opId = "
            "Job2.opId). "
            "In HTTP requests it is passed as header 'X-OPID' by the Control Service.",
        )
        config_builder.add(
            EXECUTION_ID,
            "",
            True,
            "An identifier to be associated with the current VDK execution."
            "If left empty it will be auto-generated. "
            "An instance of a running Data Job deployment is called an execution. "
            "Data Job execution can run a Data Job one or more times. "
            "It can be passed externally (in case of external re-tries are considered part of 1 execution).",
        )
        config_builder.add(
            ATTEMPT_ID,
            "",
            True,
            "An identifier to be associated with the current VDK execution attempt."
            "If left empty it will be auto-generated. "
            "An instance of a running Data Job deployment is called an execution. "
            "Data Job execution can run a Data Job one or more times."
            "Each distinct run would a single attempt.",
        )
        config_builder.add(
            TEMPORARY_WRITE_DIRECTORY,
            tempfile.gettempdir(),
            True,
            "Temporary data job write directory, to be used if job needs to"
            " write temporary files to local storage during job execution"
            " (since writing to any directory might be restricted on a"
            " deployment basis). Default value is tempfile.gettempdir()."
            " Deletion of temporary files in the default directory is managed"
            " by the underlying OS the data job executes on. There is"
            " no guarantee that files created during a local data job execution"
            " will  be present or absent for the next execution.",
        )


class EnvironmentVarsConfigPlugin:
    """
    Configuration loaded from environment variables.
    """

    def __normalize_key(self, key):
        return key.replace("-", "_").replace(".", "_").upper()

    @hookimpl(trylast=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        """
        Infer config values from environment variables.
        """
        log.info(
            "Will try to detect configuration from environment variables (using VDK_{KEY_NAME})"
        )
        description = """Attempts to load all defined configurations using environment variables by adding prefix "VDK_".
For example: if config with key "my_key" is defined, it will search for environment variables "VDK_MY_KEY"
If a key is defined with a dot it will be replaced with underscore - 'my.key' it will lookup env variable "VDK_MY_KEY"
Note: Configuration that you set with environment variables affects only local development..
Configuration for cloud execution is done in 'config.ini' file or the vdk cli.
         """
        config_builder.add(
            key="__config_provider__ environment variables",
            default_value="always_enabled",
            description=description,
        )

        config_keys = config_builder.list_config_keys()
        log.debug(
            f"Founds config keys: {config_keys}. Will check if environment variable is set for any"
        )
        upper_cased_env = {k.upper(): v for k, v in os.environ.items()}
        for key in config_keys:
            normalized_key_with_vdk_prefix = VDK_ + self.__normalize_key(key)
            normalized_key = self.__normalize_key(key)
            value = upper_cased_env.get(normalized_key)
            value_with_vdk_prefix = upper_cased_env.get(normalized_key_with_vdk_prefix)
            if value_with_vdk_prefix is not None:
                log.debug(
                    f"Found environment variable {normalized_key_with_vdk_prefix} for key {key}"
                )
                config_builder.set_value(key, value_with_vdk_prefix)
            elif value is not None:
                log.debug(f"Found environment variable {normalized_key} for key {key}")
                config_builder.set_value(key, value)


class JobConfigIniPlugin:
    def __init__(self):
        self.__job_path: Optional[pathlib.Path] = None

    @hookimpl
    def vdk_start(self, command_line_args: list) -> None:
        # TODO: improve hooks/plugin framework to enable per sub-command configuration in a less hacky way.
        if command_line_args and "run" in command_line_args:
            index = command_line_args.index("run")
            if len(command_line_args) > index + 1:
                self.__job_path = pathlib.Path(command_line_args[index + 1])

    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        description = f"""Used only during vdk run  - load configuration specified in
config.ini file in the data job's root directory.
It can be overridden by environment variables configuration (set by operators in Cloud deployment or by users locally).
         """

        config_builder.add(
            key="__config_provider__job config ini",
            default_value="always_enabled",
            description=description,
        )
        config_builder.add(
            key=PYTHON_VERSION,
            default_value=None,
            show_default_value=False,
            description="The python version, which is to be used for the data job deployment."
            " This configuration is to only be set in the [job] section of the config"
            ".ini file, and will be ignored if set in any other way (e.g., as an environment"
            " variable). As the configuration is optional, if not set, the default python version"
            " set by the Control Service would be used. To see what python versions"
            " are supported by a Versatile Data Kit deployment, use the `vdk info` command.\n"
            " EXAMPLE USAGE:\n"
            " --------------\n"
            "[job]\n"
            "schedule_cron = */50 * * * *\n"
            "python_version = 3.9",
        )

        if (
            self.__job_path
            and self.__job_path.exists()
            and self.__job_path.is_dir()
            and self.__job_path.joinpath("config.ini").exists()
        ):
            log.info("Detected config.ini. Will try to read config.ini.")

            job_config = JobConfig(self.__job_path)
            config_builder.set_value(JobConfigKeys.TEAM.value, job_config.get_team())
            config_builder.set_value(
                JobConfigKeys.TEAM, job_config.get_team()
            )  # kept for backward compatibility
            config_builder.set_value(
                JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_USER_ERROR.value,
                job_config.get_contacts_notified_on_job_failure_user_error(),
            )
            config_builder.set_value(
                JobConfigKeys.NOTIFIED_ON_JOB_SUCCESS.value,
                job_config.get_contacts_notified_on_job_success(),
            )
            config_builder.set_value(
                JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_PLATFORM_ERROR.value,
                job_config.get_contacts_notified_on_job_failure_platform_error(),
            )
            config_builder.set_value(
                JobConfigKeys.ENABLE_ATTEMPT_NOTIFICATIONS.value,
                job_config.get_enable_attempt_notifications(),
            )
            config_builder.set_value(
                JobConfigKeys.PYTHON_VERSION.value,
                job_config.get_python_version(),
            )

            for key, value in job_config.get_vdk_options().items():
                config_builder.set_value(key, value)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os

import click
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution
from vdk.api.plugin.hook_markers import GROUP_NAME
from vdk.internal import vdk_build_info
from vdk.internal.builtin_plugins.run.execution_environment import ExecutionEnvironment

try:  # importlib.metadata is used in 3.8+, importlib_metadata is used in 3.7
    from importlib import metadata
except ImportError as e:
    import importlib_metadata as metadata

log = logging.getLogger(__name__)


# Returns the version of the package named dist_name
def get_version(
    dist_name="vdk-core",
):  # Change default value if project is renamed and does not equal the setuptools metadata.name
    try:
        return get_distribution(dist_name).version
    except DistributionNotFound:  # pragma: no cover
        return "unknown"


def _dist_name(dist: metadata.PathDistribution):
    return dist.metadata["Name"]


def build_details():
    try:
        build = [
            f"{key}={value}"
            for key, value in vdk_build_info.__dict__.items()
            if not key.startswith("_")
        ]
        return ", ".join(build)
    except Exception as e:
        log.warning(f"Failed to get build details due to error: {e}")


def list_installed_plugins():
    """
    :return: List of pairs (plugin distribution name, plugin name)
    """
    dist_plugin_pairs = []
    for dist in list(metadata.distributions()):
        for ep in dist.entry_points:
            if ep.group == GROUP_NAME:
                dist_plugin_pairs.append((_dist_name(dist), ep.name))
    return dist_plugin_pairs


def list_installed_plugins_versions():
    """
    :return: a string "*plugin_name*, version *plugin_version \n *next_plugin_name*, version *next_plugin_version*" etc
    """
    dist_plugin_pairs = list_installed_plugins()

    plugin_list = [
        plugin + " (from package " + dist + ", version " + get_version(dist) + ")"
        for dist, plugin in dist_plugin_pairs
    ]
    return "\n".join(plugin_list) if len(plugin_list) > 0 else "None"


def get_version_info():
    return (
        f"Version: {get_version()}{os.linesep}"
        f"Build details: {build_details()}{os.linesep}"
        f"Python version: {ExecutionEnvironment().get_python_version()}{os.linesep}{os.linesep}"
        f"Installed plugins:{os.linesep}{list_installed_plugins_versions()}"
    )


@click.command(help="Print the version of the client.")
@click.pass_context
def version(ctx: click.Context):
    # all necessary info is printed by LogVersionInfoPlugin
    log.info(
        f"Versatile Data Kit (VDK){os.linesep}{get_version_info()}{os.linesep + '-' * 80}"
    )
    pass



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import sys
from dataclasses import dataclass
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional

import click
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core.context import CoreContext
from vdk.internal.core.statestore import StoreKey
from vdk.internal.plugin.plugin import PluginRegistry

log = logging.getLogger(__name__)


@dataclass
class ParsedCall:
    """
    Parsed Hook Call.
    """

    def __init__(self, name: str, kwargs: Dict[str, Any]) -> None:
        self.__dict__.update(kwargs)
        self._name = name

    def __repr__(self) -> str:
        """
        :return: The hook call invocation as string
        """
        d = self.__dict__.copy()
        del d["_name"]
        return f"<ParsedCall {self._name!r}(**{d!r})>"


def pluggy_result_to_string(outcome: Any) -> str:
    try:
        result = outcome.get_result()
    except BaseException as e:
        outcome.force_exception(e)
        return ""

    result_string = ""
    for item in result:
        result_string = f"\nJSON config override is {item}"

    return result_string


class HookRecorder:
    """Record all hooks invocations.

    This wraps all the hook calls/invocations, recording each call
    before propagating the normal calls.
    """

    def __init__(self, plugin_registry: PluginRegistry) -> None:
        self._plugin_registry = plugin_registry
        self.calls: List[ParsedCall] = []

        def before(hook_name: str, hook_impls: Any, kwargs: Any) -> None:
            print(
                f"------>> About to call hook {hook_name}\n"
                f"         with args:\n"
                f"         {kwargs}.\n"
                f"         Hook Impl: {hook_impls}",
                file=sys.stderr,
            )
            self.calls.append(ParsedCall(hook_name, kwargs))

        def after(outcome: Any, hook_name: str, hook_impls: Any, kwargs: Any) -> None:
            print(
                f"<<------ Finished call hook {hook_name}\n"
                f"         with args:\n"
                f"         {kwargs}.\n"
                f"         Hook Impl: {hook_impls}\n",
                f"         Outcome: {pluggy_result_to_string(outcome)}",
                file=sys.stderr,
            )

        self._undo_wrapping = (
            self._plugin_registry.plugin_manager().add_hookcall_monitoring(
                before, after
            )
        )

    def finish_recording(self) -> None:
        """
        Do not record hook calls anymore.
        """
        self._undo_wrapping()

    def get_calls(self) -> List[ParsedCall]:
        """
        Get a list of currently record calls
        """
        return [call for call in self.calls]


@click.command(help="Print hello and the CLI context.")
@click.pass_context
def hello(ctx: click.Context) -> None:
    """
    Just prints hello for testing purposes.
    """
    click.echo(f"Hello! Nice to meet you. Here are some nerdy details about me:")
    import json

    def _default_handler(o: Any) -> Any:
        from vdk.internal.core.context import CoreContext
        from vdk.internal.core.config import Configuration

        if (
            isinstance(o, click.core.Context)
            or isinstance(o, JobContext)
            or isinstance(o, CoreContext)
            or isinstance(o, Configuration)
        ):
            return vars(o)
        return str(o)

    click.echo(
        f"My context is {json.dumps(vars(ctx), default=_default_handler, indent=4, sort_keys=True)}"
    )


class DebugPlugins:
    """
    Plugin which adds some debug functionalities
    """

    def __init__(self) -> None:
        self.when = 1
        self.hook_recorder: Optional[HookRecorder] = None

    @hookimpl
    def vdk_start(
        self, plugin_registry: PluginRegistry, command_line_args: List
    ) -> None:
        """
        check if debug is needed and active it if yes.
        We also check with env variable.
        We do not use vdk_configure hook in order to trigger the recorder
        at earliest moment possible.
        """
        if "--debug-hooks" in command_line_args:
            # a bit hacky but works for hidden options
            command_line_args.remove("--debug-hooks")
            self.hook_recorder = HookRecorder(plugin_registry)
        if os.getenv("VDK_DEBUG_HOOKS_ENABLED", "no") == "yes":
            self.hook_recorder = HookRecorder(plugin_registry)

    def set_debug_hooks(self, ctx: click.Context, param: Any, value: Any):
        self.hook_recorder = HookRecorder(ctx.obj.plugin_registry)

    @hookimpl
    def vdk_initialize(self, context: CoreContext) -> None:
        plugin_registry = context.plugin_registry
        log.debug(f"Plugins loaded\n: {plugin_registry}")

    @hookimpl
    def vdk_command_line(self, root_command: click.Group) -> None:
        """
        Modify command line arguments to add debug option with callback (instead of above appraoch)
        """
        root_command.add_command(hello)
        self.add_debug_flag(root_command)

    @staticmethod
    def add_debug_flag(root_command):
        def set_debug(ctx: click.Context, param: Any, value: Any) -> None:
            if value and not ctx.resilient_parsing:
                log.debug("Enabling debug")
                core_context = cast(CoreContext, ctx.obj)
                core_context.state.set(StoreKey[bool]("vdk.debug"), True)

        debug_option = click.option(
            "--debug",
            "-D",
            type=click.BOOL,
            is_eager=True,
            expose_value=False,
            default=False,
            is_flag=True,
            callback=set_debug,
            help="Run the command in debug mode",
        )
        run_cmd = cast(click.Command, root_command.get_command(None, "run"))
        if run_cmd:
            root_command.add_command(debug_option(run_cmd))



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import List
from typing import Optional

from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.core.statestore import StoreKey


class ExecutionStateStoreKeys:
    """
    Keys used to keep data about data job execution (or run) state
    """

    JOB_NAME: StoreKey[str] = StoreKey("vdk.job_name")
    JOB_GIT_HASH: StoreKey[str] = StoreKey("vdk.job_git_hash")
    STEPS_SUCCEEDED: StoreKey[List[str]] = StoreKey("vdk.steps_succeeded")
    STEPS_STARTED: StoreKey[List[str]] = StoreKey("vdk.steps_started")
    STEPS_FAILED: StoreKey[List[str]] = StoreKey("vdk.steps_failed")
    EXECUTION_RESULT: StoreKey[Optional[ExecutionResult]] = StoreKey(
        "vdk.execution_result"
    )
    TEMPLATE_NAME: StoreKey[str] = StoreKey("vdk.template_name")



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from enum import Enum
from enum import IntEnum


class RunStatus(IntEnum):
    """
    The status of a data job run
    """

    SUCCESS = 0
    PLATFORM_ERROR = 1
    USER_ERROR = 2
    RUNNING = 3


class ExecutionStatus(str, Enum):
    """
    Just a few constants
    """

    SUCCESS = "success"
    NOT_RUNNABLE = "not_runnable"
    ERROR = "error"
    SKIP_REQUESTED = "skip_requested"



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.internal.builtin_plugins.connection.decoration_cursor import DecorationCursor
from vdk.internal.builtin_plugins.connection.decoration_cursor import ManagedOperation
from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Cursor


class RecoveryCursor(PEP249Cursor):
    """
    Extends PEP249Cursor to provide:
        * query and parameters executed
        * exception that occurred during execution
        * tooling for operation recovery

    See connection_hook_spec#db_connection_recover_operation for more details and examples how to use it.
    """

    def __init__(
        self,
        native_cursor: PEP249Cursor,
        log,
        exception,
        managed_operation: ManagedOperation,
        decoration_operation_callback,
    ):
        super().__init__(native_cursor, log)
        self.__exception = exception
        self.__managed_operation = managed_operation
        self.__decoration_operation_callback = decoration_operation_callback
        self.__retries = 0

    def get_exception(self) -> Exception:
        """
        Retrieve the original exception with which the SQL operation failed.

        :return: Exception
        """
        return self.__exception

    def get_managed_operation(self) -> ManagedOperation:
        """
        Retrieve an object that contains information about the query and query parameters used in
        the database operation. The retrieved Data Transfer Object (DTO) is purposed
        to curate the query and parameters.

        :return: ManagedOperation
            Query and parameters DTO
        """
        return self.__managed_operation

    def get_retries(self) -> int:
        """
        Fetch retries made using retry_operation().

        :return: int
            Number of operation retries performed
        """
        return self.__retries

    def execute(self, operation, parameters=None) -> None:
        """
        Execute an additional query purposed for the recovery of the original operation.

        :param operation: helper query to facilitate operation recovery
        :param parameters: helper query parameters
        """
        managed_operation = ManagedOperation(operation, parameters)
        if self.__decoration_operation_callback:
            self._log.debug("Before executing recovery query:\n%s" % operation)
            self.__decoration_operation_callback(
                decoration_cursor=DecorationCursor(
                    self._cursor, self._log, managed_operation
                )
            )
        self._log.info(
            "Executing recovery query:\n%s" % managed_operation.get_operation()
        )
        try:
            super().execute(*managed_operation.get_operation_parameters_tuple())
            self._log.info("Executing recovery query SUCCEEDED.")
        except Exception as e:
            self.retries_increment()
            self._log.warning(f"Executing recovery query FAILED. Exception: {e}")
            raise e

    def retry_operation(self) -> None:
        """
        Retry original operation to recover.
        """
        # could potentially enforce max retries here globally - in favour of per custom error handler
        self.retries_increment()
        retry_number = self.get_retries()

        self._log.info(
            f"Retrying attempt #{retry_number} "
            f"for query:\n{self.get_managed_operation().get_operation()}"
        )
        try:
            super().execute(
                *self.get_managed_operation().get_operation_parameters_tuple()
            )
            self._log.info(f"Retrying attempt #{retry_number} for query SUCCEEDED.")
        except Exception as e:
            self._log.warning(
                f"Retrying attempt #{retry_number} for query FAILED. Exception: {e}"
            )
            raise e

    def retries_increment(self) -> None:
        self.__retries += 1



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from copy import deepcopy
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from vdk.api.job_input import ISecrets
from vdk.api.plugin.plugin_input import ISecretsServiceClient

from ...core.errors import log_and_throw
from ...core.errors import ResolvableBy
from .base_secrets_impl import check_valid_secret

log = logging.getLogger(__name__)

SecretValue = Union[int, float, str, list, dict, None]


class DataJobsServiceSecrets(ISecrets):
    """
    Data Jobs Secrets implementation.
    """

    __VALID_TYPES = [int, float, str, list, dict, type(None)]

    def __init__(
        self,
        job_name: str,
        team_name: str,
        secrets_service_client: ISecretsServiceClient,
        write_preprocessors: Optional[List[ISecretsServiceClient]] = None,
    ):
        log.debug(
            f"Data Job Secrets for job {job_name} with service client: {secrets_service_client}"
        )
        self._job_name = job_name
        self._team_name = team_name
        self._secrets_service_client = secrets_service_client
        self._write_preprocessors = write_preprocessors

    def get_secret(self, name: str, default_value: SecretValue = None) -> SecretValue:
        """
        :param name: The name of the secret
        :param default_value: default value ot return if missing
        """
        secrets = self.get_all_secrets()
        if name in secrets:
            return secrets[name]
        else:
            log.warning(
                "Secret {} is not among Job secrets, returning default value: {}".format(
                    name, default_value
                )
            )
            return default_value

    def get_all_secrets(self) -> Dict[str, SecretValue]:
        """
        :return: all stored secrets
        """
        return self._secrets_service_client.read_secrets(
            self._job_name, self._team_name
        )

    def set_all_secrets(self, secrets: Dict[str, SecretValue]) -> None:
        """
        Invokes the write pre-processors if any are configured.
        Persists the passed secrets overwriting all previous properties.
        """
        if self._write_preprocessors:
            secrets = deepcopy(
                secrets
            )  # keeps the outer scope originally-referenced dictionary intact
            for client in self._write_preprocessors:
                try:
                    secrets = client.write_secrets(
                        self._job_name, self._team_name, secrets
                    )
                except Exception as e:
                    log_and_throw(
                        to_be_fixed_by=ResolvableBy.USER_ERROR,
                        log=log,
                        what_happened=f"A write pre-processor of secrets client {client} had failed.",
                        why_it_happened=f"User Error occurred. Exception was: {e}",
                        consequences="SECRETS_WRITE_PREPROCESS_SEQUENCE was interrupted, and "
                        "secrets won't be written by the SECRETS_DEFAULT_TYPE client.",
                        countermeasures="Handle the exception raised.",
                    )

        for k, v in list(secrets.items()):
            check_valid_secret(k, v, DataJobsServiceSecrets.__VALID_TYPES)  # throws

        self._secrets_service_client.write_secrets(
            self._job_name, self._team_name, secrets
        )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
The code is adapted from: https://github.com/vuolter/autoupgrade
"""
import http.client
import logging
import re
import urllib.request

import pkg_resources

log = logging.getLogger(__name__)


class Package:
    """
    PackageCheck class, holds one package.
    """

    __slots__ = ["index", "pkg", "http"]

    def __init__(self, pkg, index):
        """
        Args:
            pkg (str): name of package
            index (str): alternative index, if not given build artifactory will be used. Include
                         full index url, e.g. https://example.com/simple
        """
        self.pkg = pkg
        self.index = index.rstrip("/")
        self.http = urllib.request

    def check(self) -> bool:
        """
        Check if pkg has a later version
        Returns true if later version exists
        """
        log.debug(f"Check version for package {self.pkg} from index {self.index} ...")
        current = self._get_current()
        highest = self._get_highest_version_from_index()
        log.debug(
            f"Current version of {self.pkg} is {current} and highest is {highest}"
        )

        return highest > current

    def _get_current(self) -> tuple:
        try:
            current = self._ver_to_tuple(
                pkg_resources.get_distribution(self.pkg).version
            )
        except pkg_resources.DistributionNotFound:
            current = (-1,)
        return current

    def _get_highest_version_from_index(self) -> tuple:
        url = f"{self.index}/{self.pkg}/"

        with self.http.urlopen(url, timeout=10) as response:
            response: http.client.HTTPResponse

            if response.status >= 300:
                log.debug(
                    f"Got error status from {url}: {response.status}, {response.read()}. "
                    f"Cannot get higher version for {self.pkg}. "
                )
                return (-1,)

            pattr = rf">{self.pkg}-(.+?)<"
            versions = map(
                self._ver_to_tuple,
                re.findall(pattr, response.read().decode("utf-8"), flags=re.I),
            )
            if not versions:
                return (-1,)

            # Avoid non-stable versions like dev, rc, etc.
            versions = list(versions)
            stable_versions = [i for i in versions if len(i) < 4]

            return max(stable_versions)

    @staticmethod
    def _ver_to_tuple(value):
        """
        Convert version like string to a tuple of integers.
        """
        return tuple(int(_f) for _f in re.split(r"\D+", value) if _f)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.ingestion import ingester_configuration
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core.config import ConfigurationBuilder


class IngesterConfigurationPlugin:
    """
    Configuration plugin to ensure all plugin configurations, relevant to ingestion
    plugins are ready when the plugins are loaded.
    """

    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        """
        Here we define the configuration settings needed for all ingestion
        plugins with reasonable defaults.
        """
        # Plugin-related configurations
        config_builder.add(
            key="INGEST_METHOD_DEFAULT",
            default_value=None,
            description="""Default Ingestion method to be used. It is
            possible to specify a coma-separated string of ingestion
            plugins, and the payload would be ingested by all plugins
            specified.
            Example:
                INGEST_METHOD_DEFAULT="http"
                and
                INGEST_METHOD_DEFAULT="http,file"
                are both acceptable.
            """,
        )
        config_builder.add(
            key="INGEST_TARGET_DEFAULT",
            default_value=None,
            description="Default Ingestion target to be used.",
        )
        config_builder.add(
            key="INGEST_PAYLOAD_PREPROCESS_SEQUENCE",
            default_value=None,
            description="""A string of coma-separated ingestion pre-processing
            plugin method names, indicating the sequence in which the
            ingestion pre-processing plugins would process the payload. For
            example:
                   INGEST_PAYLOAD_PREPROCESS_SEQUENCE="pre-ingest-process,
                   ingest-pre-process"
            would mean that the payload sent for ingestion would be first
            processed by a `pre-ingest-process` plugin, then by the
            `ingest-pre-process` plugin, and finally would be ingested by
            the ingestion plugin specified by the 'method' argument of the
            send_object_for_ingestion()/send_tabular_data_for_ingestion()
            methods, or specified by the `INGEST_METHOD_DEFAULT` environment
            variable.
            NOTE: If an ingestion plugin implements both pre-processing and
            ingestion logic, it would need to be specified both in
            INGEST_PAYLOAD_PREPROCESS_SEQUENCE and INGEST_METHOD_DEFAULT.
            """,
        )
        config_builder.add(
            key="INGEST_PAYLOAD_POSTPROCESS_SEQUENCE",
            default_value=None,
            description="""A string of coma-separated ingestion post-
            processing plugin method names, indicating the sequence, in which
            the ingestion post-processing plugins would process the metadata
            generated during the ingestion.
            Example:
                    INGEST_PAYLOAD_POSTPROCESS_SEQUENCE="post-ingest-process,
                    post-process"
            The above example shows how the ingestion post-processing plugins
            could be specified. In the example, the metadata generated from
            the ingestion process would be processed first by the
            `post-ingest-process`, and then by the `post-process` plugins.
            NOTE: If an ingestion plugin implements pre-processing, ingestion
            and post-processing logic, it would need to be specified in the
            INGEST_PAYLOAD_PREPROCESS_SEQUENCE, INGEST_METHOD_DEFAULT and
            INGEST_PAYLOAD_POSTPROCESS_SEQUENCE environment variables.
            """,
        )

        # Configure ingestion specific environment variables
        ingester_configuration.add_definitions(config_builder=config_builder)

    @hookimpl
    def finalize_job(self, context: JobContext) -> None:
        context.ingester.close()



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.connection.decoration_cursor import DecorationCursor
from vdk.internal.builtin_plugins.connection.decoration_cursor import ManagedOperation
from vdk.internal.builtin_plugins.connection.recovery_cursor import RecoveryCursor
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core.context import CoreContext
from vdk.internal.core.statestore import CommonStoreKeys


class QueryDecoratorPlugin:
    """
    Prefix each DB query with job name and operation ID
    """

    @hookimpl
    def initialize_job(self, context: JobContext) -> None:
        self._job_name = context.name

    @hookimpl
    def vdk_initialize(self, context: CoreContext) -> None:
        self._op_id = context.state.get(CommonStoreKeys.OP_ID)

    @hookimpl
    def db_connection_decorate_operation(
        self, decoration_cursor: DecorationCursor
    ) -> None:
        managed_operation = decoration_cursor.get_managed_operation()
        managed_operation.set_operation(
            "\n".join(
                ["-- job_name: {job_name}", "-- op_id: {op_id}", "{operation}"]
            ).format(
                job_name=self._job_name,
                op_id=self._op_id,
                operation=managed_operation.get_operation(),
            )
        )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging

from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.config.vdk_config import LOG_CONFIG
from vdk.internal.builtin_plugins.termination_message import (
    writer_configuration,
)
from vdk.internal.builtin_plugins.termination_message.file_util import (
    WriteToFileAction,
)
from vdk.internal.builtin_plugins.termination_message.writer_configuration import (
    WriterConfiguration,
)
from vdk.internal.builtin_plugins.version.version import get_version
from vdk.internal.core import errors
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext

log = logging.getLogger(__name__)


class TerminationMessageWriterPlugin:
    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder):
        writer_configuration.add_definitions(config_builder)

    @hookimpl
    def vdk_exit(self, context: CoreContext, exit_code: int):
        self.write_termination_message(
            errors.get_blamee_overall(),
            errors.get_blamee_overall_user_error(),
            context.configuration,
        )

    def write_termination_message(
        self,
        error_overall=None,
        user_error=None,
        configuration=None,
        execution_skipped=False,
    ):
        termination_message_writer_cfg = WriterConfiguration(configuration)

        try:
            termination_message_output_file = (
                termination_message_writer_cfg.get_output_file()
            )

            if (
                termination_message_writer_cfg.get_writer_enabled()
                and termination_message_output_file
            ):
                # Hide termination messages when running locally
                show_termination_log_messages = (
                    configuration.get_value(LOG_CONFIG).lower() != "local"
                )
                file_util = WriteToFileAction(
                    termination_message_output_file, show_termination_log_messages
                )

                if show_termination_log_messages:
                    log.debug(
                        f'Writing termination message to file "{termination_message_output_file}"'
                    )
                self._execute_termination_action(
                    file_util, error_overall, user_error, execution_skipped
                )
        except Exception as e:
            log.exception(f"Failed to write termination message. See exception: {e}")

    @staticmethod
    def _execute_termination_action(
        file_util, error_overall, user_error, execution_skipped
    ):
        termination_message = {"vdk_version": get_version()}

        if execution_skipped:
            status = "Skipped"
        elif not error_overall:
            status = "Success"
        elif user_error:
            status = "User error"
        else:
            status = "Platform error"

        termination_message["status"] = status
        file_util.write_to_file(json.dumps(termination_message))



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Any

from vdk.internal.builtin_plugins.connection.decoration_cursor import ManagedOperation
from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Cursor
from vdk.internal.builtin_plugins.connection.proxy_cursor import ProxyCursor


class ExecuteOperationResult:
    def __init__(self, native_result: Any):
        self.__native_result = native_result

    def get_native_result(self) -> Any:
        """
        Returns the result of PEP249Cursor.execute invocation using the native cursor.
        Since this is vendor specific the result type is any - but usually it's number of selected or updated rows.
        :return: Any
        """
        return self.__native_result


class ExecutionCursor(ProxyCursor):
    """
    Extends ProxyCursor to provide:
        * ability to directly access and execute operations with the native cursor.
        Generally it is not execpted to be used. Perhaps only if default implementation does not work (which should be almost never)
        But it's useful to be exposed in case the native library used provide some extra features.
        For example Impala (impyla driver) provides a way to get the profile of a query after it's executed
        which can be printed to aid debugging.
        * Give access to the result of execute operation (through) ExecuteOperationResult which is vendor specific.

    See connection_hook_spec#db_connection_execute_operation for more details and examples how to use it.
    """

    def __init__(
        self,
        native_cursor: PEP249Cursor,
        managed_operation: ManagedOperation,
        log=logging.getLogger(__name__),
    ):
        super().__init__(native_cursor, log)
        self.__managed_operation = managed_operation
        self.__native_cursor = native_cursor

    def get_managed_operation(self) -> ManagedOperation:
        """
        Retrieve an object that contains information about the query and query parameters used in
        the database operation. The retrieved Data Transfer Object (DTO) is purposed
        to curate the query and parameters.

        :return: ManagedOperation
            Query and parameters DTO
        """
        return self.__managed_operation



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import os
from pathlib import Path

import click
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.connection.managed_connection_base import (
    ManagedConnectionBase,
)
from vdk.internal.builtin_plugins.run.standalone_data_job import (
    StandaloneDataJobFactory,
)
from vdk.internal.util.decorators import closing_noexcept_on_close

OUTPUT_OPTIONS = ["json", "text"]

# The command is hidden as it's still experimental.
# TODOs:
# Remove logs  by default only show output (and perhaps warning logs)
# When logging omit "job details" logs
# Make clear how it is authenticated.


@click.command(
    name="sql-query",
    hidden=True,  # Hidden as it is experimental
    help="executes SQL query against configured database. "
    "The database used is configured with 'db_default_type' option. "
    "See vdk config-help for more info. ",
)
@click.option(
    "-q", "--query", type=click.STRING, required=True, help="The query to be executed"
)
@click.option(
    "-o",
    "--output",
    type=click.Choice(OUTPUT_OPTIONS, False),
    default="text",
    help=f"Output format. It can be one of: {OUTPUT_OPTIONS}.",
)
@click.pass_context
def sql_query(ctx: click.Context, query: str, output: str) -> None:
    # TODO: os.getcwd() is passed to StandaloneDataJob, this means that if the current directory is same as data job
    # it would authenticate using the data job credentials (if there are any like keytab)
    # Is that a good idea?

    with StandaloneDataJobFactory.create(
        data_job_directory=Path(os.getcwd())
    ) as job_input:
        conn: ManagedConnectionBase = job_input.get_managed_connection()
        with closing_noexcept_on_close(conn.cursor()) as cursor:
            cursor.execute(query)
            column_names = (
                [column_info[0] for column_info in cursor.description]
                if cursor.description
                else ()  # same as the default value for the headers parameter of the tabulate function
            )
            # TODO: this is basically emulating
            # https://github.com/vmware/versatile-data-kit/blob/main/projects/vdk-control-cli/src/vdk/internal/control/utils/output_printer.py
            # It will be best printer functionality to be reusable (either moved to vdk-core or in a library/plugin)
            result = cursor.fetchall()
            if output.lower() == "json":
                result = [
                    {column_names[i]: v for i, v in enumerate(row)} for row in result
                ]
                click.echo(json.dumps(result))
            elif output.lower() == "text":
                from tabulate import tabulate

                click.echo(tabulate(result, headers=column_names))
            else:
                raise ValueError(
                    f"Unsupported output format. Choose between: {OUTPUT_OPTIONS}"
                )


class QueryCommandPlugin:
    @staticmethod
    @hookimpl
    def vdk_command_line(root_command: click.Group) -> None:
        root_command.add_command(sql_query)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import pathlib
from dataclasses import dataclass
from datetime import datetime
from typing import cast

from vdk.api.job_input import IJobArguments
from vdk.api.plugin.core_hook_spec import JobRunHookSpecs
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.builtin_plugins.run.execution_results import StepResult
from vdk.internal.builtin_plugins.run.execution_state import ExecutionStateStoreKeys
from vdk.internal.builtin_plugins.run.file_based_step import JobFilesLocator
from vdk.internal.builtin_plugins.run.file_based_step import StepFuncFactory
from vdk.internal.builtin_plugins.run.file_based_step import TYPE_PYTHON
from vdk.internal.builtin_plugins.run.file_based_step import TYPE_SQL
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.builtin_plugins.run.job_input_error_classifier import whom_to_blame
from vdk.internal.builtin_plugins.run.run_status import ExecutionStatus
from vdk.internal.builtin_plugins.run.step import Step
from vdk.internal.core import errors
from vdk.internal.core.context import CoreContext
from vdk.internal.core.errors import SkipRemainingStepsException
from vdk.internal.core.statestore import CommonStoreKeys

log = logging.getLogger(__name__)


@dataclass
class JobArguments(IJobArguments):
    arguments: dict

    def get_arguments(self) -> dict:
        return self.arguments


class DataJobFactory:
    @staticmethod
    def new_datajob(
        data_job_directory: pathlib.Path | None,
        core_context: CoreContext,
        name: str | None = None,
    ) -> DataJob:
        """
        Create a new data job
        :param data_job_directory: the source code of the data job that will be executed
        :param core_context: Core context of the CLI. Upon run , the data job will span child context to keep its context/state
        :param name: the name of the job. Leave it out and it will be infered from the directory name.
        """
        return DataJob(data_job_directory, core_context, name)


class DataJobDefaultHookImplPlugin:
    """
    Default implementation of main plugin hooks in Data Job Run Cycle.
    Plugins may decorate or replace some of the implementations
    """

    @staticmethod
    @hookimpl(trylast=True)
    def run_step(context: JobContext, step: Step) -> StepResult:
        start_time = datetime.utcnow()
        exception = None
        details = None
        blamee = None

        try:
            log.debug(f"Processing step {step.name} ...")
            step_executed = step.runner_func(step, context.job_input)
            log.debug("Processing step %s completed successfully" % step.name)
            status = (
                ExecutionStatus.SUCCESS
                if step_executed
                else ExecutionStatus.NOT_RUNNABLE
            )
        except SkipRemainingStepsException as e:
            status = ExecutionStatus.SKIP_REQUESTED
            details = errors.MSG_WHY_FROM_EXCEPTION(e)
            log.debug(
                f"Job execution skipped from step: {step.name}. Because skip_remaining_steps() method "
                + "was invoked"
            )
        except Exception as e:
            status = ExecutionStatus.ERROR
            details = errors.MSG_WHY_FROM_EXCEPTION(e)
            blamee = whom_to_blame(e, __file__, context.job_directory)
            exception = e
            errors.log_exception(
                blamee,
                log,
                what_happened=f"Processing step {step.name} completed with error.",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                consequences="I will not process the remaining steps (if any), "
                "and this Data Job execution will be marked as failed.",
                countermeasures="See exception and fix the root cause, so that the exception does "
                "not appear anymore.",
                exception=e,
            )

        return StepResult(
            name=step.name,
            type=step.type,
            start_time=start_time,
            end_time=datetime.utcnow(),
            status=status,
            details=details,
            exception=exception,
            blamee=blamee,
        )

    @staticmethod
    @hookimpl(trylast=True)
    def run_job(context: JobContext) -> ExecutionResult:
        """The script that runs the actual run of the data job.
        It executes the provided steps starting from context.steps in sequential order
        """
        start_time = datetime.utcnow()
        exception = None
        blamee = None
        steps = context.step_builder.get_steps()
        step_results = []

        if len(steps) == 0:
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=log,
                what_happened="Data Job execution has failed.",
                why_it_happened="Data Job has no steps.",
                consequences="Data job execution will not continue.",
                countermeasures="Please include at least 1 valid step in your Data Job. Also make sure you are passing the correct data job directory.",
            )

        execution_status = ExecutionStatus.SUCCESS
        for current_step in steps:
            step_start_time = datetime.utcnow()
            try:
                res = context.core_context.plugin_registry.hook().run_step(
                    context=context, step=current_step
                )
            except BaseException as e:
                blamee = whom_to_blame(e, __file__, context.job_directory)
                exception = e
                errors.log_exception(
                    blamee,
                    log,
                    what_happened=f"Processing step {current_step.name} completed with error.",
                    why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                    consequences="I will not process the remaining steps (if any), "
                    "and this Data Job execution will be marked as failed.",
                    countermeasures="See exception and fix the root cause, so that the exception does "
                    "not appear anymore.",
                    exception=e,
                )
                res = StepResult(
                    name=current_step.name,
                    type=current_step.type,
                    start_time=step_start_time,
                    end_time=datetime.utcnow(),
                    status=ExecutionStatus.ERROR,
                    details=errors.MSG_WHY_FROM_EXCEPTION(e),
                    exception=e,
                    blamee=blamee,
                )

            step_results.append(res)
            # errors.clear_intermediate_errors()  # step completed successfully, so we can forget errors
            if res.status == ExecutionStatus.ERROR:
                execution_status = ExecutionStatus.ERROR
                break
            if res.status == ExecutionStatus.SKIP_REQUESTED:
                # We keep the status as Success, but we skip all remaining steps
                break

        execution_result = ExecutionResult(
            context.name,
            context.core_context.state.get(CommonStoreKeys.EXECUTION_ID),
            start_time,
            datetime.utcnow(),
            execution_status,
            step_results,
            exception,
            blamee,
        )
        return execution_result

    @staticmethod
    @hookimpl
    def initialize_job(context: JobContext):
        # TODO: consider split into collect_steps hooks for better clarity and ease.
        # though let's first gather some data on how useful such new hooks would be.
        if context.job_directory is None:
            log.info(
                "Data Job directory is not specified. Default job initialization will be skipped."
            )
            return

        file_locator: JobFilesLocator = JobFilesLocator()
        script_files = file_locator.get_script_files(context.job_directory)

        for file_path in script_files:
            if file_path.name.lower().endswith(".sql"):
                step = Step(
                    name=file_path.name,
                    type=TYPE_SQL,
                    runner_func=StepFuncFactory.run_sql_step,
                    file_path=file_path,
                    job_dir=context.job_directory,
                )
            elif file_path.name.lower().endswith(".py"):
                # TODO: check for run method.
                step = Step(
                    name=file_path.name,
                    type=TYPE_PYTHON,
                    runner_func=StepFuncFactory.run_python_step,
                    file_path=file_path,
                    job_dir=context.job_directory,
                )
            else:
                log.info("Skipping file as it is not a valid job step: %s" % file_path)
                continue
            context.step_builder.add_step(step)

    @staticmethod
    @hookimpl
    def finalize_job(context: JobContext):
        pass


class DataJob:
    """
    Object representing a data job.
    Data Job is abstraction representing definition of a data job set by data engineer.
    Data Job is a sequence of steps executed in a certain order.
    Data JOb run cycle is encapsulated by run method.

    Prefer to use DataJobFactory to create new data job.
    """

    def __init__(
        self,
        data_job_directory: pathlib.Path | None,
        core_context: CoreContext,
        name: str | None = None,
    ):
        if data_job_directory is None and name is None:
            raise ValueError(
                "Cannot initialize DataJob. "
                "At least one of data job directory or data job name need to be set. "
            )
        self._name = data_job_directory.name if name is None else name
        self._data_job_directory = data_job_directory
        """
        We need to create child context which will contain only for this job execution.
        This is since we want to have multiple job executions within same process (for example templates executions)
        """
        self._core_context = core_context.create_child_context()
        self._plugin_hook = cast(
            JobRunHookSpecs, self._core_context.plugin_registry.hook()
        )

    @property
    def name(self):
        return self._name

    # TODO: this also can be a hook - e.g job run_cycle_algorithm
    def run(
        self, args: dict = None, template_name: str | None = None
    ) -> ExecutionResult:
        """
        This is basic implementation of Data Job run(execution) cycle algorithm.
        All stages are pluggable as hooks.
         * Initialize - Initialize job's main functionalities -e.g database connections,logging,collecting steps, etc.
         * Run Job -Takes care of starting and running a data job
         * Run steps - run the actual steps. This is where hte user code will be invoked.
         * Finalize - after the job finishes, do any finalization -clean up, send monitoring, etc.
        """
        if args is None:
            args = {}
        if template_name:
            self._core_context.state.set(
                ExecutionStateStoreKeys.TEMPLATE_NAME, template_name
            )

        if not self._core_context.plugin_registry.has_plugin(
            DataJobDefaultHookImplPlugin.__name__
        ):
            self._core_context.plugin_registry.load_plugin_with_hooks_impl(
                DataJobDefaultHookImplPlugin(), DataJobDefaultHookImplPlugin.__name__
            )

        from vdk.internal.builtin_plugins.templates.template_impl import TemplatesImpl

        job_context = JobContext(
            name=self._name,
            job_directory=self._data_job_directory,
            core_context=self._core_context,
            job_args=JobArguments(args),
            templates=TemplatesImpl(
                job_name=self.name,
                core_context=self._core_context,
                template_name=template_name,
            ),
        )
        self._plugin_hook.initialize_job(context=job_context)

        start_time = datetime.utcnow()
        try:
            return self._plugin_hook.run_job(context=job_context)
        except BaseException as ex:
            blamee = whom_to_blame(ex, __file__, job_context.job_directory)
            errors.log_exception(
                blamee,
                log,
                what_happened=f"Data Job {self._name} completed with error.",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(ex),
                consequences="I will not process the remaining steps (if any), "
                "and this Data Job execution will be marked as failed.",
                countermeasures="See exception and fix the root cause, so that the exception does "
                "not appear anymore.",
                exception=ex,
            )
            execution_result = ExecutionResult(
                self._name,
                self._core_context.state.get(CommonStoreKeys.EXECUTION_ID),
                start_time,
                datetime.utcnow(),
                ExecutionStatus.ERROR,
                [],
                ex,
                blamee,
            )
            return execution_result

        finally:  # TODO: we should pass execution result to finalize_job somehow ...
            self._plugin_hook.finalize_job(context=job_context)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Ingestion Utilities
"""
import datetime
import itertools
import logging
import queue
import threading
import uuid
from decimal import Decimal
from json import JSONEncoder
from typing import Any
from typing import List

from vdk.internal.core import errors

log = logging.getLogger(__name__)


class AtomicCounter:
    """
    Thread-safe incrementing counter
    """

    def __init__(self, start: int = 0):
        """Init a new atomic counter to start value"""
        self.value: int = start
        self._lock = threading.Lock()

    def increment(self, step: int = 1) -> int:
        """Atomically increment the counter by step, return new value"""
        self.get_and_increment(step)
        return self.value

    def get_and_increment(self, step: int = 1) -> int:
        """Atomically increment counter by step, return old value"""
        with self._lock:
            old_value: int = self.value
            self.value += step
            return old_value

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return str(self)


class DecimalJsonEncoder(JSONEncoder):
    """
    This class is used to avoid an issue with the __verify_payload_format serialization check.
    Normally, including data of type Decimal would cause that check to fail so we've amended
    the default JsonEncoder object used to convert Decimal values to floats to avoid this issue.
    """

    def default(self, obj):
        if isinstance(obj, Decimal):
            return float(obj)
        return super().default(obj)


def get_page_generator(data, page_size=10000):
    # TODO add support for Pandas
    try:
        while True:
            page = data.fetchmany(page_size)
            if not page:
                return
            yield page
    except AttributeError:
        it = iter(data)
        while True:
            page = list(itertools.islice(it, page_size))
            if not page:
                return
            yield page


def validate_column_count(data: iter, column_names: iter):
    if data:
        if len(column_names) != len(data[0]):
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=log,
                what_happened="Failed to post tabular data for ingestion.",
                why_it_happened="The number of column names are not matching the number of values in at least on of"
                "the rows. You provided columns: '{column_names}' and data row: "
                "'{data_row}'".format(column_names=column_names, data_row=data[0]),
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures="Check the data and the column names you are providing, their count should match.",
            )


def convert_table(table: iter, column_names: iter) -> List[dict]:
    """
    Converts tabular data into dictionary objects

    :param table: iter
       A representation of a two-dimensional array that allows iteration over rows.
    :param column_names: iter
       Names of the table columns.
    :return: list of dicts containing the converted table objects.
    """
    converted_rows = []
    for row in table:
        cdf_row = dict()
        for index, value in enumerate(row):
            value = _handle_special_types(value)
            cdf_row[column_names[index]] = value
        converted_rows.append(cdf_row)

    return converted_rows


def _handle_special_types(value: Any) -> Any:
    """
    Handle data types that require special care to be correctly processed.

    :param value: Any
        The value to be converted. Only `datetime.date` and `uuid.UUID` objects
        are converted. All other data types are returned as is.
    """
    if isinstance(value, datetime.date):
        return value.isoformat()
    if isinstance(value, uuid.UUID):
        return str(
            value
        )  # there is no default json serializer for UUID type, hence cast to string
    return value


def wait_completion(objects_queue: queue.Queue, payloads_queue: queue.Queue):
    objects_queue.join()
    payloads_queue.join()


def is_iterable(obj: Any) -> bool:
    """
    According to Python doc this is the most reliable way to determine if object is iterable.
    See https://docs.python.org/3/library/collections.abc.html#collections.abc.Iterable
    :param obj: the object to check
    :return: true or false :)
    """
    try:
        _ = iter(obj)
        return True
    except TypeError:
        return False



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import re
from collections import OrderedDict
from textwrap import wrap
from typing import cast

import click
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.core.context import CoreContext

log = logging.getLogger(__name__)

CONFIG_HELP = """
Configuring VDK is done via:
{config_providers}
-----------

List of configuration variables:
{config_variables}

------------

How to set them?\n
The above listed configuration variables can be set using following configuration providers:\n
{config_providers}
"""


def description_line_wrapping(description, description_start=40, description_length=60):
    wrapped_text_list = []
    description_paragraphs = description.splitlines()
    for paragraph in description_paragraphs:
        wrapped_text_list += wrap(
            paragraph,
            width=description_length,
            break_long_words=False,
            break_on_hyphens=False,
        )
    indent = " " * description_start
    # we do not need to wrap the first line (it's already padded to the config key)
    for index, line in enumerate(wrapped_text_list[1:], start=1):
        wrapped_text_list[index] = indent + wrapped_text_list[index]
    return "\n".join(wrapped_text_list)


def description_cleanup(config_variable: str, description: str, description_start: int):
    if description:
        description = re.sub(" +", " ", description)
        description = re.sub("\n ", "\n", description)
        description = description.strip()
        if len(config_variable) > description_start:
            description = "  " + description
    return description


# Description length can be configured by changing the default value for description_length
def formatted_configuration_help(
    config_variable, description, description_start=40, description_length=60
):
    config_variable = config_variable.strip()
    padding = " " * (description_start - len(config_variable))
    conf_help = "{config_var}{padding}{description}"
    description = description_cleanup(config_variable, description, description_start)
    return conf_help.format(
        config_var=config_variable,
        padding=padding,
        description=description_line_wrapping(
            description, description_start, description_length
        ),
    )


def generate_config_list_help(variable_to_description_map):
    config_list = "\n\n"
    conf_variable_list = []
    for key in variable_to_description_map:
        conf_variable_list.append(
            formatted_configuration_help(key, variable_to_description_map[key])
        )
    return config_list.join(conf_variable_list)


@click.command(
    help="Print details on configuration options of vdk. "
    "It includes all configuration options added by plugins."
    """

Examples:

\b
# if we call config-help we get similar output to one below.
# first we provide details on how those configuration can be set (e.g environment variables, files, etc)
# Next list of all possible configuration variables supported.
vdk config-help

\b
Configuring VDK is done via:
environment variables                   Attempts to load all defined configurations using
                                        environment variables by adding prefix "VDK_".
                                        ...
-----
List of configuration variables:
DB_DEFAULT_TYPE                         Default DB connection provided by VDK when ..
EXECUTION_ID                            An identifier to be associated with the current VDK run ...
    """
)
@click.pass_context
def config_help(ctx: click.Context) -> None:
    configuration = cast(CoreContext, ctx.obj).configuration

    vars_to_descriptions = {}
    providers_descriptions = {}
    for k in configuration.list_config_keys():
        description = configuration.get_description(k)
        if description:
            if k.startswith("__config_provider__"):
                name = k[len("__config_provider__") :].strip()
                providers_descriptions[name] = description
            else:
                vars_to_descriptions[k] = description

    click.echo(
        CONFIG_HELP.format(
            config_providers=generate_config_list_help(
                OrderedDict(sorted(providers_descriptions.items()))
            ),
            config_variables=generate_config_list_help(
                OrderedDict(sorted(vars_to_descriptions.items()))
            ),
        )
    )


class ConfigHelpPlugin:
    """
    The following plugins add a new command (vdk conflig-help) which would enable
    users to query for possible configurations. Each plugin declare their possible configuration
    in vdk_configure hook (using config_builder.add) providng descriptions and other attributes,
    Those then can be queried by end users. See vdk_configure doc for more information.
    """

    @hookimpl
    def vdk_command_line(self, root_command: click.Group) -> None:
        """
        Modify command line arguments to add config help option
        """
        root_command.add_command(config_help)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import Container
from typing import Optional
from typing import Tuple

from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Cursor


class ManagedOperation:
    """
    An object that contains information about the query and query parameters used in the database operation.
    A Data Transfer Object or DTO.
    """

    def __init__(self, operation: str, parameters: Optional[Container]):
        self.__operation = operation
        self.__parameters = parameters
        self.__operation_decorated = operation
        self.__parameters_decorated = parameters

    def get_initial_operation(self) -> str:
        """
        The initial operation (usually SQL expression) before being modified by any plugin.
        :return: str
        """
        return self.__operation

    def get_initial_parameters(self) -> Optional[Container]:
        """
        The initial parameters before being modified by any plugin.
        :return: Optional[Container]
        """
        return self.__parameters

    def get_operation(self) -> str:
        """
        The operation (usually SQL expression) that is expected to be executed.
        It may have been modified by other plugins.
        :return: str
        """
        return self.__operation_decorated

    def get_parameters(self) -> Optional[Container]:
        """
        The parameters that are expected to be used during operation execution.
        It may have been modified by other plugins.
        :return: Optional[Container]
        """
        return self.__parameters_decorated

    def get_operation_parameters_tuple(self) -> Tuple[str, Optional[Container]]:
        """
        The operation (usually SQL expression) and parameters that is expected to be executed.
        They may have been modified by other plugins.
        :return: Tuple[str, Optional[Container]]
        """
        return self.__operation_decorated, self.__parameters_decorated

    # maybe track modification history
    def set_operation(self, decorated_operation: str):
        """
        Set the active operation that will be executed. Use it to decorate or change the operation (or SQL expression).
        :param decorated_operation: str
            The operation modified
        :return:
        """
        self.__operation_decorated = decorated_operation

    def set_parameters(self, decorated_parameters: Optional[Container]):
        """
        Set the active parameters that will be used during operation execution.
        Use it to decorate or change the parameters.
        :param decorated_parameters: Optional[Container]
            The parameters modified
        :return:
        """
        self.__parameters_decorated = decorated_parameters

    def set_operation_parameters_tuple(
        self, decorated_operation: str, decorated_parameters: Optional[Container]
    ):
        """
        Set the active operation and parameters that will be executed.
        Use it to decorate or change the operation (or SQL expression) and parameters.
        :param decorated_operation: str
        :param decorated_parameters: Optional[Container]
        :return:
        """
        self.set_operation(decorated_operation)
        self.set_parameters(decorated_parameters)


class DecorationCursor(PEP249Cursor):
    """
    Extends PEP249Cursor to provide:
        * query and parameters to be executed
        * tooling for curating the operation
        * tooling for pre-executing other decorative queries
    """

    def __init__(
        self, native_cursor: PEP249Cursor, log, managed_operation: ManagedOperation
    ):
        super().__init__(native_cursor, log)
        self.__managed_operation = managed_operation

    def get_managed_operation(self) -> ManagedOperation:
        """
        Retrieve operation DTO to curate the query and parameters.

        :return: ManagedOperation
            Query and parameters DTO
        """
        return self.__managed_operation

    def execute(self, operation, parameters=None) -> None:
        """
        Execute an additional query purposed for decoration of the original operation.

        :param operation: decoration query
        :param parameters: decoration query parameters
        """
        self._log.info(f"Executing decoration query:\n{operation}")
        try:
            super().execute(operation, parameters)
            self._log.info("Executing decoration query SUCCEEDED.")
        except Exception as e:
            self._log.warning(f"Executing decoration query FAILED. Exception: {e}")
            raise e



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import configparser
import fileinput
import logging
import os
import pathlib
import re
import sys
from configparser import ConfigParser
from configparser import MissingSectionHeaderError
from enum import Enum
from typing import Dict
from typing import List
from typing import Union

from vdk.internal.core.config import convert_value_to_type_of_default_type
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import VdkConfigurationError


log = logging.getLogger(__name__)


class JobConfigKeys(str, Enum):
    TEAM = "team"
    SCHEDULE_CRON = "schedule_cron"
    PYTHON_VERSION = "python_version"
    ENABLE_EXECUTION_NOTIFICATIONS = "enable_execution_notifications"
    NOTIFICATION_DELAY_PERIOD_MINUTES = "notification_delay_period_minutes"
    NOTIFIED_ON_JOB_FAILURE_USER_ERROR = "notified_on_job_failure_user_error"
    NOTIFIED_ON_JOB_FAILURE_PLATFORM_ERROR = "notified_on_job_failure_platform_error"
    NOTIFIED_ON_JOB_SUCCESS = "notified_on_job_success"
    NOTIFIED_ON_JOB_DEPLOY = "notified_on_job_deploy"
    ENABLE_ATTEMPT_NOTIFICATIONS = "enable_attempt_notifications"
    DB_DEFAULT_TYPE = "db_default_type"


class JobConfig:
    """
    User facing configuration of a Data Job.
    For more see the user wiki
    """

    def __init__(
        self, data_job_path: Union[pathlib.Path, str], should_fail_missing_config=False
    ):
        self._config_ini = configparser.ConfigParser()
        self._config_file = os.path.join(data_job_path, "config.ini")
        self._should_fail_missing_config = should_fail_missing_config

        if not os.path.isfile(self._config_file):
            if should_fail_missing_config:
                raise VdkConfigurationError(
                    ErrorMessage(
                        summary="Error while loading config.ini file",
                        what="Cannot extract job Configuration",
                        why=f"Configuration file config.ini is missing in data job path: {data_job_path}",
                        consequences="Cannot deploy and configure the data job without config.ini file.",
                        countermeasures="config.ini must be in the root of the data job folder. "
                        "Make sure the file is created "
                        "or double check the data job path is passed correctly.",
                    )
                )
            else:
                log.info("Missing config.ini file.")

        self._read_config_ini_file(
            config_parser=self._config_ini, configuration_file_path=self._config_file
        )

    def get_team(self) -> str:
        """
        Specified which is the team that owns the data job.

        Teams is a way to group data jobs that belong together and are managed by same people.
        It will control who has permissions to manages the data job.
        """
        return self._get_value("owner", JobConfigKeys.TEAM.value)

    def set_team_if_exists(self, value):
        """
        If 'team' option exists in section 'owner' of config.ini, value param is assigned to it and
        config.ini file is overwritten with the new team value.
        Returns True if team is found and successfully set to 'value' in config.ini, False - otherwise
        """
        return self._set_value("owner", JobConfigKeys.TEAM.value, value)

    def get_schedule_cron(self) -> str:
        return self._get_value("job", JobConfigKeys.SCHEDULE_CRON.value)

    def get_python_version(self) -> str:
        return str(self._get_value("job", JobConfigKeys.PYTHON_VERSION.value))

    def get_enable_execution_notifications(self) -> bool:
        return self._get_boolean(
            "contacts",
            JobConfigKeys.ENABLE_EXECUTION_NOTIFICATIONS.value,
            fallback=True,
        )

    def get_notification_delay_period_minutes(self) -> int:
        return self._get_positive_int(
            "contacts",
            JobConfigKeys.NOTIFICATION_DELAY_PERIOD_MINUTES.value,
            fallback=240,
        )

    def get_contacts_notified_on_job_failure_user_error(self) -> List[str]:
        """
        List of email addresses to be notified on job execution failure caused by user code or user configuration problem
        """
        return self._get_contacts(
            JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_USER_ERROR.value
        )

    def get_contacts_notified_on_job_failure_platform_error(self) -> List[str]:
        """
        List of email addresses to be notified on job execution failure caused by a platform problem.
        """
        return self._get_contacts(
            JobConfigKeys.NOTIFIED_ON_JOB_FAILURE_PLATFORM_ERROR.value
        )

    def get_contacts_notified_on_job_success(self) -> List[str]:
        """
        List of email addresses to be notified on job execution success.
        """
        return self._get_contacts(JobConfigKeys.NOTIFIED_ON_JOB_SUCCESS.value)

    def get_contacts_notified_on_job_deploy(self) -> List[str]:
        """
        List of email addresses to be notified of job deployment outcome.
        """
        return self._get_contacts(JobConfigKeys.NOTIFIED_ON_JOB_DEPLOY.value)

    def get_enable_attempt_notifications(self) -> bool:
        """
        Flag to enable or disable the email notifications on a single attempt
        (in other words each automatic retry will send an email).
        """
        enable_attempt_notif_value = self._get_value(
            "contacts", JobConfigKeys.ENABLE_ATTEMPT_NOTIFICATIONS.value
        )
        if (
            enable_attempt_notif_value == ""
        ):  # if the parameter hasn't been set in config.ini, we default to False
            return False
        else:
            return convert_value_to_type_of_default_type(
                JobConfigKeys.ENABLE_ATTEMPT_NOTIFICATIONS.value,
                enable_attempt_notif_value,
                False,
            )

    def get_db_type(self) -> str:
        """
        The default database type used by the job (IMPALA, PRESTO)
        """
        return self._get_value("job", JobConfigKeys.DB_DEFAULT_TYPE.value)

    def get_vdk_options(self) -> Dict[str, str]:
        if self._config_ini.has_section("vdk"):
            return dict(self._config_ini["vdk"])
        else:
            return {}

    def _get_value(self, section, key):
        if self._config_ini.has_option(section, key):
            return self._config_ini.get(section, key)
        return ""

    def _set_value(self, section, key, value) -> bool:
        success = False
        if self._config_ini.has_option(section, key):
            for line in fileinput.input(self._config_file, inplace=1):
                if line.replace(" ", "").startswith(f"{key}="):
                    success = True
                    line = f"{key} = {value}\n"
                sys.stdout.write(line)
        return success

    def _get_contacts(self, key):
        contacts_str = self._get_value("contacts", key).strip()
        contacts = []
        if contacts_str:
            contacts = [x.strip() for x in re.split("[;,]", contacts_str)]
        return contacts

    def _get_boolean(self, section, key, fallback=None) -> bool:
        return self._config_ini.getboolean(section, key, fallback=fallback)

    def _get_positive_int(self, section, key, fallback=None) -> int:
        try:
            value = self._config_ini.getint(section, key, fallback=fallback)
            if value <= 0:
                raise ValueError()
            return value
        except ValueError:
            raise VdkConfigurationError(
                ErrorMessage(
                    summary="Invalid configuration property.",
                    what=f"The configuration '{key}' property in the job's config.ini file is not valid.",
                    why=f"The '{key}' configuration should be a positive integer, "
                    f"but instead '{self._get_value(section, key)}' is found.",
                    consequences="Cannot configure the data job without valid configuration.",
                    countermeasures=f"Change the value of the '{key}' property in the job's config.ini file to "
                    f"a positive integer and redeploy the job.",
                )
            )

    @staticmethod
    def _read_config_ini_file(
        config_parser: ConfigParser, configuration_file_path: str
    ) -> None:
        """
        Read the Data Job config.ini file.

        :param config_parser: ConfigParser instance to be used for reading the
        configuration file.
        :param configuration_file_path: Path of the config.ini file
        """
        try:
            config_parser.read(configuration_file_path)
        except (MissingSectionHeaderError, Exception) as e:
            log.debug(e, exc_info=True)  # Log the traceback in DEBUG mode.
            raise VdkConfigurationError(
                ErrorMessage(
                    summary="Error while parsing config file.",
                    what="Cannot parse the Data Job configuration file"
                    f" {configuration_file_path}.",
                    why=f"Configuration file config.ini is probably corrupted. Error: {e}",
                    consequences="Cannot deploy and configure the data job "
                    "without "
                    " properly set config.ini file.",
                    countermeasures="config.ini must be UTF-8 compliant. "
                    "Make sure the file does not contain special "
                    "Unicode characters, or that your text editor "
                    "has not added such characters somewhere in the file.",
                )
            )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import functools
import logging
from contextlib import contextmanager
from typing import TypeVar

log = logging.getLogger(__name__)

T = TypeVar("T")


@contextmanager
def closing_noexcept_on_close(thing: T) -> T:
    try:
        yield thing
    finally:
        try:
            thing.close()
        except Exception as e:
            log.info(
                f"Failed to close {thing}. This is not an issue. Will ignore it and continue. Exception was '{e}'."
            )
            pass


class LogDecorator:
    def __init__(self, logger, message=None):
        self.logger = logger
        self.message = message

    def __call__(self, fn):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            try:
                message = self.message if self.message else fn.__qualname__
                self.logger.debug(f"Starting: {message} ({args},{kwargs})")
                result = fn(*args, **kwargs)
                self.logger.debug(f"Finished: {message}.")
                return result
            except Exception as ex:
                self.logger.debug(f"Exception {ex}")
                raise

        return decorated



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import types
from typing import Any

from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Cursor


class ProxyCursor(PEP249Cursor):
    """
    PEP249 Proxy cursor.
    It wraps all methods and proxies (non-override methods) to the passed native cursor.
    Allowing the wrapped cursor to be used as a native cursor as well for native (vendor specific) methods.

    """

    def __init__(self, cursor: Any, log: logging.Logger = logging.getLogger(__name__)):
        super().__init__(cursor, log)

    def __getattr__(self, attr):
        """
        Dynamic interception and il of any (non-overridden) attribute access.
        In case an attribute is not explicitly managed (customized by overriding e.g. execute()) -
        this attribute is looked up then the call is delegated, ensuring default behaviour success path.

        First, the non-managed attribute call is redirected to the wrapped native cursor if attribute available,
        otherwise to the superclass if attribute is present.
        If the attribute is not specified by both the native cursor nor the superclass, an AttributeError is raised.

        Default behaviour availability unblocks various ManagedCursor usages that rely on
        currently not explicitly defined in the scope of ManagedCursor attributes.
        E.g. SQLAlchemy dependency that uses the ManagedCursor, does require some specific attributes available.

        For more details on customizing attributes access, see PEP562.
        """
        # native cursor
        if hasattr(self._cursor, attr):
            if isinstance(getattr(self._cursor, attr), types.MethodType):

                def method(*args, **kwargs):
                    return getattr(self._cursor, attr)(*args, **kwargs)

                return method
            return getattr(self._cursor, attr)
        # superclass
        if hasattr(super(), attr):
            if isinstance(getattr(super(), attr), types.MethodType):

                def method(*args, **kwargs):
                    return getattr(super(), attr)(*args, **kwargs)

                return method
            return getattr(super(), attr)
        raise AttributeError



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Dict

log = logging.getLogger(__name__)


class SqlArgumentSubstitutor:
    """
    Substitute text in provided SQL from provided dictory
    """

    def __init__(self, dictionary: Dict):
        self.dictionary = dictionary

    def substitute(self, sql: str) -> str:
        """Substitute text in provided SQL.

        e.g. a SQL like
        "SELECT {col1}, {col2} FROM {table}" and dictionary like {'col1': 'id', 'col2': 'pa__processed_ts', 'table': 'vm'}
        would result in the following SQL:
        "SELECT id, pa__processed_ts FROM vm"

        The substitution is case-sensitive.
        In case of variable that is used in the text, but not defined in the dictionary, there will be NO substitution
        and no error. For example:
        "SELECT {col1}, {col2} FROM {table}" and dictionary like {'col1': 'xxxxxx'} will result in
        "SELECT xxxxxx, {col2} FROM {table}"

        :return: String containing all the given text with substituted variables.
        """

        for key in self.dictionary:
            property_name = "{" + key + "}"
            if sql.find(property_name) != -1:
                property_value = str(self.dictionary[key])
                sql = sql.replace(property_name, property_value)
        return sql



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from abc import ABCMeta
from abc import abstractmethod
from typing import List
from typing import Tuple

import pluggy


class PluginHookRelay:
    """
    Helper class to relay method execution of underlying hook.
    It's merely to enable autocomplete and reading documentation of hook spec when using it.
    For each declared function spec, add method with same name and implementation:
    `return self.pm.hook.method_name`

    When invoking the hook note that return of the hook result is an array of all results (for all plugins hooks)
    unless hook is marked with firstresult=True

    """

    def __init__(self, hook_relay):
        self.__hook_relay = hook_relay

    def __dir__(self):
        return dir(self.__hook_relay)

    def __getattr__(self, key):
        hook = getattr(self.__hook_relay, key)
        self.__dict__[key] = hook
        return hook


class PluginException(Exception):
    """
    Exception thrown if error happens when working with plugins.
    """

    pass


"""
Alias for the type of plugin hook call result returned in hookWrapper=True types of plugin hooks
"""
HookCallResult = (
    pluggy.callers._Result if pluggy.__version__ < "1.0" else pluggy._callers._Result
)


class IPluginRegistry(metaclass=ABCMeta):
    """
    Plugins are basically python modules that implement one or more plugin hooks with some kind of specification.
    Specification is a method name and signature that must be implemented and is decorated with hook_markers.hookspec

    For example we can have hookspec defined this way:

    .. highlight:: python
    .. code-block:: python
        class DataJobRunHookSpec
            @hookspec
            def configure_job(self, context: JobContext, configuration: Config):
                pass

    Create a new plugin hook is simply implementing a method with the same name and marked with hook_markers.hookimpl.

    .. highlight:: python
    .. code-block:: python
        @hookimpl()
        def configure_job(self, context: JobContext, configuration: Config):
            ...

    """

    @abstractmethod
    def list_plugins(self) -> List[Tuple[str, str]]:
        """
        :return: pairs of plugin name and plguin module/class which holds the plugin hook implementations
        """

    @abstractmethod
    def has_plugin(self, name: str) -> bool:
        """
        :param name: the name of the plugin
        :return: Returns true if a plugin has been loaded.
        """
        pass

    @abstractmethod
    def is_plugin_load_success(self) -> bool:
        """
        Whether plugin loading suceeded. load_plugin_with_hooks_impl method
        finished successfully
        :return: True if plugin load was successful
        """
        pass

    @abstractmethod
    def load_plugin_with_hooks_impl(
        self, module_or_class_with_hook_impls: object, name: str
    ) -> None:
        """
        Load new plugin with hook implementations

        :param module_or_class_with_hook_impls: can be module (python package) or class or class instance
        :param name: the name of the plugin. If None it will be the canonical name of the module/class passed.
        which contain the hook implementations
        """
        pass

    @abstractmethod
    def add_hook_specs(self, module_or_class_with_hookspecs: object) -> None:
        """
        Add hook specifications. They must be annotated with hook_markers.hookimpl.
        :param module_or_class_with_hookspecs:
        :return:
        """
        pass

    @abstractmethod
    def hook(self) -> PluginHookRelay:
        """
        Returns hook relay use to execute the plugin hooks.
        If there is plugin hook with spec: vdk_initialize(context: JobContext)
        Then all its implementations will be invoked with hook().vdk_initialize(context=job_context)

        Note that hook must be invoked using keyword arguments.

        :return: Hook relay which to use to execute the plugin hooks with
        """
        pass



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.job_secrets import secrets_config
from vdk.internal.builtin_plugins.job_secrets.inmemsecrets import (
    InMemSecretsServiceClient,
)
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.core.config import ConfigurationBuilder


class SecretsApiPlugin:
    """
    Define the basic configuration needed for Secrets API.
    """

    @hookimpl(tryfirst=True)
    def vdk_configure(self, config_builder: ConfigurationBuilder) -> None:
        secrets_config.add_definitions(config_builder)

    @hookimpl
    def initialize_job(self, context: JobContext) -> None:
        context.secrets.set_secrets_factory_method("memory", InMemSecretsServiceClient)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0


def check_valid_secret(k: str, v: str, supported_types=None) -> None:
    """
    Check if secret key and value are valid
    """

    if supported_types is None:
        supported_types = []

    if str != type(k) or k.strip() != k or "".join(k.split()) != k:
        msg = (
            f"Secret {k} is of unsupported type or has unsupported name. "
            f"Only string secrets with no whitespaces in the name are supported."
        )
        raise ValueError(msg)

    if not supported_types:
        supported_types = [int, float, str, list, type(None)]

    if type(v) not in supported_types:
        msg = (
            f"Value for secret  {k} is of unsupported type {type(v)}. "
            f"Only int, float, str, list, and NoneType types are supported. "
        )
        raise ValueError(msg)



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass
from dataclasses import field
from typing import Any

from vdk.internal.core.errors import VdkConfigurationError

# Consider ConfigValue should be primitive type perhaps? and not just any object
ConfigValue = Any
ConfigKey = str

log = logging.getLogger(__name__)


def convert_value_to_type_of_default_type(
    key: ConfigKey, v: ConfigValue, default_value: ConfigValue
) -> ConfigValue:
    """
    Allows for configurations to be converted to the type of their default value.

    E.g. IMPALA_PORT is converted to int, because it's default value is int.
    """
    if default_value is not None:
        if type(default_value) == bool and type(v) != bool:
            allowed_values = ["true", "false", "1", "0", "yes", "no", "y", "n"]
            if str(v).lower() not in allowed_values:
                msg = (
                    f"Provided configuration "
                    f'"{key}={v}" is invalid. Allowed values for {key} are {allowed_values}'
                )
                raise VdkConfigurationError(msg)
            v = str(v).lower() in ["true", "1", "yes", "y"]
        else:
            try:
                v = type(default_value)(v)  # cast to type of default_value:
            except ValueError:
                msg = (
                    f'Provided configuration "{key}={v}" is invalid. '
                    f'Cannot cast "{v}" to {str(type(default_value).__name__)}'
                )
                raise VdkConfigurationError(msg)
    return v


def _normalize_config_key(key: ConfigKey):
    return str(key).lower()


@dataclass(frozen=True)
class Configuration:
    """
    The configuration of the application. It's immutable.

    Use ConfigurationBuilder to build it.
    """

    __config_key_to_description: dict[ConfigKey, str]
    __config_key_to_value: dict[ConfigKey, ConfigValue]
    __config_key_to_default_value: dict[ConfigKey, ConfigValue] = field(
        default_factory=dict
    )
    __config_key_to_sensitive: dict[ConfigKey, bool] = field(default_factory=dict)

    def __getitem__(self, key: ConfigKey):
        key = _normalize_config_key(key)
        return self.get_value(key)

    def get_value(self, key: ConfigKey) -> ConfigValue:
        """
        Return configuration value associated with a given key or None if it does not exists.

        For required configuration see get_required_value

        :param key: the configuration key (e.g db_host, service_uri, etc.)
        :return: the value corresponding to the configuration key
        """
        key = _normalize_config_key(key)
        default_value = self.__config_key_to_default_value.get(key)
        value = self.__config_key_to_value.get(key, default_value)
        return value

    def get_required_value(self, key: ConfigKey) -> ConfigValue:
        """
        Return configuration value associated with a given key or throws VdkConfigurationError if it's missing.

        Use get_value if configuration is not required

        :param key: the configuration key (e.g db_host, service_uri, etc.)
        :return: the value corresponding to the configuration key
        :raises VdkConfigurationError
        """
        key = _normalize_config_key(key)
        value = self.get_value(key)
        if value is None:
            raise VdkConfigurationError(
                f"Required configuration {key} is missing."
                f"This will cause configuration error and some functionality may not work."
                f"Please provide correct configuration value. "
                f"See help for how to configure the tool"
            )
        return value

    def get_description(self, key: ConfigKey) -> str | None:
        """
        Get description of hte configuration with that key. It will be used to be printed in help.

        :param key: the config key
        :return: description
        """
        key = _normalize_config_key(key)
        return self.__config_key_to_description.get(key)

    def is_sensitive(self, key: ConfigKey) -> bool | None:
        """
        Check if config key is sensitive, e.g. password

        :param key: the config key
        :return: True if sensitive, else False
        """
        return self.__config_key_to_sensitive.get(key)

    def list_config_keys(self) -> list[ConfigKey]:
        """
        List all added (defined) config keys

        :return: list of key names.
        """
        return [k for k in self.__config_key_to_default_value.keys()]


@dataclass()
class ConfigurationBuilder:
    """
    Builder used to configure the app.

    Usually a plugin would define what variables it needs using add method.
    While other plugins will populate them using set_value method (e.g environment variable plugin. yaml file plugin, etc.).
    """

    __config_key_to_description: dict[ConfigKey, str]
    __config_key_to_value: dict[ConfigKey, ConfigValue]
    __config_key_to_default_value: dict[ConfigKey, ConfigValue]
    __config_key_to_sensitive: dict[ConfigKey, bool]

    def __init__(self):
        self.__config_key_to_description = dict()
        self.__config_key_to_default_value = dict()
        self.__config_key_to_value = dict()
        self.__config_key_to_sensitive = dict()

    def add(
        self,
        key: ConfigKey,
        default_value: ConfigValue,
        show_default_value=True,
        description=None,
        is_sensitive=False,
    ) -> ConfigurationBuilder:
        """
        Add new configuration variable definition.

        :param key: The configuration key. If already exist variable will be updated.
        :param default_value: Default value for the configuration Can be None.
        :param show_default_value: default value will appear in help as well.
        The default value type will enforce the type of the option. Can be None - in this case the type would str.
        :param description: Set description if you want config variable to appear in command line help .
        :param is_sensitive: Set to True if the configuration variable represents sensitive data, e.g. password.
        False by default. Appends 'This option is marked as sensitive' to the description if set to True
        It is strongly recommended to set description. If no description is set the config key will be hidden.
        TODO: in the future we should require description always and have separate hidden=True/False instead
        :return: self so it can be chained like builder.add(..).set_value(...)...
        """
        key = _normalize_config_key(key)
        self.__config_key_to_default_value[key] = default_value
        self.__config_key_to_sensitive[key] = is_sensitive
        if description and show_default_value:
            self.__add_public(key, description, is_sensitive, default_value)
        elif description:
            self.__add_public(key, description, is_sensitive)
        self.__adjust_type_if_value_set(key, default_value)
        return self

    def set_value(self, key: ConfigKey, value: ConfigValue) -> ConfigurationBuilder:
        """
        This will set or update the value for a configuration with a given key.

        It will try to cast it to the specified default type inferred from default value (set with #add method)
        :param key: the configuration key
        :param value: the configuration value.
        :return: self so it can be chained like builder.set_value(..).add(...)...
        """
        key = _normalize_config_key(key)
        default_value = self.__config_key_to_default_value.get(key)
        self.__config_key_to_value[key] = convert_value_to_type_of_default_type(
            key, value, default_value
        )
        return self

    def list_config_keys(self) -> list[ConfigKey]:
        """
        List all added (defined) config keys

        :return: list of key names.
        """
        return [k for k in self.__config_key_to_default_value.keys()]

    def __add_public(
        self,
        key: ConfigKey,
        description: str,
        is_sensitive: bool = False,
        default_value: ConfigValue = None,
    ) -> None:
        if not isinstance(description, str):
            log.warning(
                f"Description for key {key} is not of type string. Converting to type string."
            )
            description = str(description)

        if default_value is not None:
            description += "\nDefault value is: '%s'." % default_value
        if is_sensitive:
            description += "\nThis option is marked as sensitive."
        self.__config_key_to_description[key] = description

    def __adjust_type_if_value_set(self, key: ConfigKey, default_value: ConfigValue):
        """
        As set_value can be called before add (which defines the configuration) we may need to adjust
        the type of the value if it's already set.
        """
        if default_value is not None and key in self.__config_key_to_value:
            self.__config_key_to_value[key] = convert_value_to_type_of_default_type(
                key, self.__config_key_to_value[key], default_value
            )

    def build(self) -> Configuration:
        """
        :return: immutable version of the Configuration
        """
        return Configuration(
            self.__config_key_to_description,
            self.__config_key_to_value,
            self.__config_key_to_default_value,
            self.__config_key_to_sensitive,
        )



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import functools
import getpass
import logging
import platform
import socket
import sys

log = logging.getLogger(__name__)


class ExecutionEnvironment:
    """
    Information about execution environment - e.g. OS, user, ip, etc.
    """

    def _get_result_noexcept(self, method, property_name):
        try:
            value = method()
        except Exception as e:
            log.debug(
                f"Failed getting {property_name}. Will use unknown instead. Exception: {e} "
            )
            value = "unknown-" + property_name
        return value

    @functools.lru_cache(maxsize=None)
    def get_machine_info(self):
        """
        Get information about the the machine being run on
        Though it's generally better to be split into multiple attributes for telemetry
        But we are putting together not to pollute table schema with too many system columns
        """
        user = self._get_result_noexcept(getpass.getuser, "user")
        ip = self._get_result_noexcept(
            lambda: socket.gethostbyname(socket.gethostname()), "ip"
        )
        os_info = self._get_result_noexcept(platform.platform, "os-info")

        return f"{user}@{ip}({os_info})"

    @functools.lru_cache(maxsize=None)
    def get_python_version(self):
        """
        Return python version being used to run the job - for example "3.7.9 64bit"
        """
        py_version = self._get_result_noexcept(
            platform.python_version, "python-version"
        )
        py_arch = self._get_result_noexcept(
            lambda: platform.architecture()[0], "python-arch"
        )

        return f"{py_version} {py_arch} ({sys.executable})"



# Copyright 2021-2023 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import inspect
from logging import Logger
from typing import Any
from typing import List
from typing import Optional

from vdk.internal.core import errors
from vdk.internal.core.config import Configuration
from vdk.internal.core.errors import ResolvableBy


def class_fqname(py_object: Any) -> str:
    """
    :param py_object: Any python object
    :return: the full qualified name of the class (package.package.classname)
    """
    module = inspect.getmodule(py_object)
    if module is None:
        return py_object.__class__.__name__
    return f"{py_object.__class__.__module__}.{py_object.__class__.__name__}"


def parse_config_sequence(
    cfg: Configuration, key: str, sep: Optional[str] = None
) -> List:
    """
    Parse a configuration variable, whose value is a string sequence, by a
    specified delimiter.
    :param cfg: Configuration
        The configuration of Versatile Data Kit.
    :param key: string
        The name of the configuration variable that is to be parsed.
    :param sep: string
        Optional. The delimiter by which the configuration variable string is
        to be split. If not specified or set as None, any whitespace will be
        considered a delimiter, and the result from the split will have no
        empty strings.
    :return:
    """
    sequence = cfg.get_value(key)
    if sequence:
        sequence = [i.strip() for i in sequence.split(sep)]
    return sequence if sequence else []


def log_plugin_load_fail(
    user_error: ResolvableBy, log: Logger, exception: Exception, group_name
):
    """
    Logs errors during plugin load. Calls the errors.log_exception() method
    which will also update the resolvable context of the VDK run.

    :param user_error: If this is a platform error or no
    :param log: The logger which will log the exception
    :param exception: The exception
    :param group_name
    :return:
    """
    errors.log_exception(
        user_error,
        log,
        f"Cannot load plugin from setuptools entrypoint for group {group_name}",
        "See exception for possible reason",
        "The CLI tool will likely abort.",
        "Re-try again. Check exception message and possibly uninstall a bad"
        " plugin (pip uninstall) Or see what plugins are installed (use `pip"
        " list` command) and if there aren't issues. Or try to reinstall the"
        " app in a new clean environment. Try to revert to previous version of"
        " the CLI tool. If nothing works open a SRE ticket.",
        exception,
    )
